"""
Tests for pcmci.py, including tests for run_pc_stable, run_mci, and run_pcmci.
"""
from __future__ import print_function
from collections import Counter, defaultdict
import itertools
import numpy as np
# from nose.tools import assert_equal
from numpy.testing import assert_equal
import pytest

from tigramite.pcmci import PCMCI
from tigramite.independence_tests.parcorr import ParCorr
from tigramite.independence_tests.oracle_conditional_independence import OracleCI
import tigramite.data_processing as pp
from tigramite.toymodels import structural_causal_processes as toys

# Pylint settings
# pylint: disable=redefined-outer-name

# Define the verbosity at the global scope
VERBOSITY = 1

# CONVENIENCE FUNCTIONS ########################################################
def assert_graphs_equal(actual, expected):
    """
    Check whether lists in dict have the same elements.
    This ignore the order of the elements in the list.
    """
    for j in list(expected):
        assert_equal(Counter(iter(actual[j])), Counter(iter(expected[j])))

def _get_parent_graph(parents_neighbors_coeffs, exclude=None):
    """
    Iterates through the input parent-neighghbour coefficient dictionary to
    return only parent relations (i.e. where tau != 0)
    """
    graph = defaultdict(list)
    for j, i, tau, _ in toys._iter_coeffs(parents_neighbors_coeffs):
        if tau != 0 and (i, tau) != exclude:
            graph[j].append((i, tau))
    return dict(graph)

def _select_links(link_ids, true_parents):
    """
    Select links given from the true parents dictionary
    """
    if link_ids is None:
        return None
    return {par : {true_parents[par][link]:'-->'} for par in true_parents \
                                            for link in link_ids}

def _get_parents_from_results(pcmci, results):
    """
    Select the significant parents from the MCI-like results at a given
    alpha_level
    """
    significant_parents = \
        pcmci.return_parents_dict(graph=results['graph'],
                                val_matrix=results['val_matrix'])
    return significant_parents

def gen_data_frame(links_coeffs, time, seed_val):
    # Set the random seed
    np.random.seed(seed_val)
    # Generate the data
    data, _ = toys.var_process(links_coeffs, T=time)
    # Get the true parents
    true_parents = _get_parent_graph(links_coeffs)
    return pp.DataFrame(data), true_parents


def gen_confounder_data_frame_mixed(links_coeffs, time, seed_val):
    # Set the random seed
    random_state = np.random.default_rng(seed_val)
    data = np.zeros((time, 3))
    data[:, 2] = random_state.binomial(n=1, p=0.5, size=time)
    for t in range(2, time):
        data[t, 0] = links_coeffs[0][0][1] * data[t + links_coeffs[0][0][0][1], 1] + random_state.normal(
            0.2 + data[t + links_coeffs[0][1][0][1], 2] * links_coeffs[0][1][1], 1)
        data[t, 1] = links_coeffs[1][0][1] * data[t + links_coeffs[1][0][0][1], 2] + random_state.normal(
            0.2 + data[t + links_coeffs[1][1][0][1], 2] * links_coeffs[1][1][1], 1)

    data_type = np.zeros(data.shape, dtype='int')
    # X2 is continuous, encoded as 1 in data_type
    data_type[:, 2] = 1

    dataframe = pp.DataFrame(data,
                             data_type=data_type)

    true_parents = _get_parent_graph(links_coeffs)
    return dataframe, true_parents


def gen_chain_data_frame_mixed(links_coeffs, time, seed_val):
    # Set the random seed
    np.random.seed(seed_val)
    # Generate the data
    data, _ = toys.var_process(links_coeffs, T=time)
    data_type = np.zeros(data.shape)
    # Get the true parents
    true_parents = _get_parent_graph(links_coeffs)
    return pp.DataFrame(data, data_type=data_type), true_parents


# TEST LINK GENERATION #########################################################
def a_chain(auto_corr, coeff, length=3):
    """
    Generate a simple chain process with the given auto-correlations and
    parents with the given coefficient strength.  A length can also be defined
    to get a longer chain.

    Parameters
    ----------
    auto_corr: float
        Autocorrelation strength for all nodes
    coeff : float
        Parent strength for all relations
    length : int
        Length of the chain
    """
    return_links = dict()
    return_links[0] = [((0, -1), auto_corr)]
    for lnk in range(1, length):
        return_links[lnk] = [((lnk, -1), auto_corr), ((lnk-1, -1), coeff)]
    return return_links

def mixed_confounder(auto_corr, coeff, length=3):
    """
    Generate a simple confounder process with the given auto-correlations and
    parents with the given coefficient strength.

    Parameters
    ----------
    auto_corr: float
        Autocorrelation strength for all nodes
    coeff : float
        Parent strength for all relations
    length : int
        Length of the confounder model.
    """
    return_links = dict()
    return_links[2] = []
    for lnk in range(0, length - 1):
        return_links[lnk] = [((lnk, -1), auto_corr), ((2, -1), coeff)]

    return return_links

# TODO implement common_driver: return two variables commonly driven by N common
# drivers which are random noise, autocorrelation as parameter
# TODO implement independent drivers, autocorrelated noise
# TODO check common_driver, independent driver cases for current variable sets
# TODO implement USER_INPUT dictionary,
# USER_INPUT = dict()


# TEST DATA GENERATION #########################################################
@pytest.fixture(params=[
    # Generate a test data sample
    # Parameterize the sample by setting the autocorrelation value, coefficient
    # value, total time length, and random seed to different numbers
    # links_coeffs,               time,  seed_val
    (a_chain(0.1, 0.9),           1000,  2),
    (a_chain(0.5, 0.6),           1000,  11),
    (a_chain(0.5, 0.6, length=5), 10000, 42)])
def a_sample(request):
    # Set the parameters
    links_coeffs, time, seed_val = request.param
    # Generate the dataframe
    return gen_data_frame(links_coeffs, time, seed_val)

# PCMCI CONSTRUCTION ###########################################################
@pytest.fixture(params=[
    # Keep parameters common for all the run_ algorithms here
    # tau_min, tau_max,  sel_link,
     (1,       2,        None),
     # (1,       2,        [0])
     ])
def a_common_params(request):
    # Return the requested parameters
    return request.param

@pytest.fixture()
    # Parameterize and return the independence test.
    # Currently just a wrapper for ParCorr, but is extendable
def a_test(request):
    return ParCorr(verbosity=VERBOSITY)

@pytest.fixture(params=[None])
    # Fixture to build and return a parameterized PCMCI.  Different selected
    # variables can be defined here.
def a_pcmci(a_sample, a_test, a_common_params, request):
    # Unpack the test data and true parent graph
    dataframe, true_parents = a_sample
    # Unpack the common parameters
    tau_min, tau_max, sel_link = a_common_params
    # Get the parameters from this request
    select_vars = request.param
    # Build the PCMCI instance
    pcmci = PCMCI(dataframe=dataframe,
                  cond_ind_test=a_test,
                  verbosity=VERBOSITY)
    # Select the correct links if they are given
    select_links = _select_links(sel_link, true_parents)
    # print(select_links)
    # Ensure we change the true parents to be the same as the selected links
    if select_links is not None:
        true_parents = select_links
    # Return the constructed PCMCI, expected results, and common parameters
    return pcmci, true_parents, tau_min, tau_max, select_links

# PC_STABLE TESTING ############################################################
@pytest.fixture(params=[
    # Keep parameters for the pc_stable algorithm here
    # pc_alpha,  max_conds_dim,  max_comb, save_iterations
     (None,      None,           3,        False),
     (0.05,      None,           1,        False),
     (0.05,      None,           10,       False),
     # (0.05,      None,           1,        True),
     # (0.05,      3,              1,        False)
    ])
def a_pc_stable_params(request):
    # Return the parameters for the pc_stable test
    return request.param

@pytest.fixture()
def a_run_pc_stable(a_pcmci, a_pc_stable_params):
    # Unpack the pcmci, true parents, and common parameters
    pcmci, true_parents, tau_min, tau_max, select_links = a_pcmci
    # Unpack the pc_stable parameters
    pc_alpha, max_conds_dim, max_combinations, save_iter = a_pc_stable_params
    # Run PC stable
    pcmci.run_pc_stable(link_assumptions=None,
                        tau_min=tau_min,
                        tau_max=tau_max,
                        save_iterations=save_iter,
                        pc_alpha=pc_alpha,
                        max_conds_dim=max_conds_dim,
                        max_combinations=max_combinations)
    # Return the calculated and expected results
    return pcmci.all_parents, true_parents

def test_pc_stable(a_run_pc_stable):
    """
    Test the pc_stable algorithm and check it calculates the correct parents.
    """
    # Unpack the calculated and true parents
    parents, true_parents = a_run_pc_stable
    # Ensure they are the same
    assert_graphs_equal(parents, true_parents)

# MCI TESTING ##################################################################
@pytest.fixture(params=[
    # Keep parameters for the mci algorithm here
    # alpha_level, max_conds_px, max_conds_py
     (0.01,        None,         None)])
def a_mci_params(request):
    # Return the parameters for the mci test
    return request.param

@pytest.fixture()
def a_run_mci(a_pcmci, a_mci_params):
    # Unpack the pcmci and the true parents, and common parameters
    pcmci, true_parents, tau_min, tau_max, select_links = a_pcmci
    # Unpack the MCI parameters
    alpha_level, max_conds_px, max_conds_py = a_mci_params
    # Run the MCI algorithm with the given parameters
    results = pcmci.run_mci(link_assumptions=None,
                            tau_min=tau_min,
                            tau_max=tau_max,
                            parents=true_parents,
                            max_conds_py=max_conds_px,
                            max_conds_px=max_conds_py,
                            alpha_level=alpha_level)
    # Return the calculated and expected results
    return _get_parents_from_results(pcmci, results), true_parents

def test_mci(a_run_mci):
    """
    Test the mci algorithm and check it calculates the correct parents.
    """
    # Unpack the calculated and true parents
    parents, true_parents = a_run_mci
    # Ensure they are the same
    assert_graphs_equal(parents, true_parents)

# PCMCI TESTING ################################################################
@pytest.fixture()
def a_run_pcmci(a_pcmci, a_pc_stable_params, a_mci_params):
    # Unpack the pcmci and the true parents, and common parameters
    pcmci, true_parents, tau_min, tau_max, select_links = a_pcmci
    # Unpack the pc_stable parameters
    pc_alpha, max_conds_dim, max_combinations, save_iter = a_pc_stable_params
    # Unpack the MCI parameters
    alpha_level, max_conds_px, max_conds_py = a_mci_params
    # Run the PCMCI algorithm with the given parameters
    results = pcmci.run_pcmci(link_assumptions=None,
                              tau_min=tau_min,
                              tau_max=tau_max,
                              save_iterations=save_iter,
                              pc_alpha=pc_alpha,
                              max_conds_dim=max_conds_dim,
                              max_combinations=max_combinations,
                              max_conds_px=max_conds_px,
                              max_conds_py=max_conds_py,
                              alpha_level=alpha_level)
    # Return the results and the expected result
    return _get_parents_from_results(pcmci, results), true_parents

def test_pcmci(a_run_pcmci):
    """
    Test the pcmci algorithm and check it calculates the correct parents.
    """
    # Unpack the calculated and true parents
    parents, true_parents = a_run_pcmci
    # Ensure they are the same
    assert_graphs_equal(parents, true_parents)

# PCMCIplus TESTING
# ################################################################
def lin_f(x): return x

def setup_nodes(auto_coeff, N):
    link_coeffs = {}
    for j in range(N):
       link_coeffs[j] = [((j, -1), auto_coeff, lin_f)]
    return link_coeffs

def a_collider(auto_coeff, coeff):
    link_coeffs = setup_nodes(auto_coeff, N=3)
    for i in [0, 2]:
        link_coeffs[1].append(((i, 0), coeff, lin_f))
    return link_coeffs

def a_rule1(auto_coeff, coeff):
    link_coeffs = setup_nodes(auto_coeff, N=3)
    link_coeffs[1].append(((0, -1), coeff, lin_f))
    link_coeffs[2].append(((1, 0), coeff, lin_f))
    return link_coeffs

def a_rule2(auto_coeff, coeff):
    link_coeffs = setup_nodes(auto_coeff, N=3)
    link_coeffs[1].append(((1, -1), coeff, lin_f))
    link_coeffs[1].append(((0, 0), coeff, lin_f))
    link_coeffs[2].append(((1, 0), coeff, lin_f))
    link_coeffs[2].append(((0, 0), coeff, lin_f))
    return link_coeffs

def a_rule3(auto_coeff, coeff):
    link_coeffs = setup_nodes(auto_coeff, N=4)
    link_coeffs[1].append(((1, -1), coeff, lin_f))
    link_coeffs[0].append(((1, 0), coeff, lin_f))
    link_coeffs[2].append(((1, 0), coeff, lin_f))
    link_coeffs[2].append(((0, 0), coeff, lin_f))
    link_coeffs[3].append(((0, 0), coeff, lin_f))
    link_coeffs[2].append(((3, 0), coeff, lin_f))

    return link_coeffs

def a_random_process(
    N, L, 
    coupling_coeffs, 
    coupling_funcs, 
    auto_coeffs, 
    tau_max, 
    contemp_fraction=0.,
    num_trials=1000,
    model_seed=None):
    """Generates random structural causal process.
       
       TODO: documentation.

    Returns
    -------
    links : dict
        Dictionary of form {0:[((0, -1), coeff, func), ...], 1:[...], ...}.
    """
    
    def lin(x): return x

    random_state = np.random.RandomState(model_seed)

    # print links
    a_len = len(auto_coeffs)
    if type(coupling_coeffs) == float:
        coupling_coeffs = [coupling_coeffs]
    c_len  = len(coupling_coeffs)
    func_len = len(coupling_funcs)

    if contemp_fraction > 0.:
        contemp = True
        L_lagged = int((1.-contemp_fraction)*L)
        L_contemp = L - L_lagged
        if L==1: 
            # Randomly assign a lagged or contemp link
            L_lagged = random_state.randint(0,2)
            L_contemp = int(L_lagged == False)

    else:
        contemp = False
        L_lagged = L
        L_contemp = 0

    # for ir in range(num_trials):
    # Random order
    causal_order = list(random_state.permutation(N))

    links = dict([(i, []) for i in range(N)])

    # Generate auto-dependencies at lag 1
    for i in causal_order:
        a = auto_coeffs[random_state.randint(0, a_len)]

        if a != 0.:
            links[i].append(((int(i), -1), float(a), lin))

    chosen_links = []
    # Create contemporaneous DAG
    contemp_links = []
    for l in range(L_contemp):

        cause = random_state.choice(causal_order[:-1])
        effect = random_state.choice(causal_order)
        while (causal_order.index(cause) >= causal_order.index(effect)
             or (cause, effect) in chosen_links):
            cause = random_state.choice(causal_order[:-1])
            effect = random_state.choice(causal_order)
        
        contemp_links.append((cause, effect))
        chosen_links.append((cause, effect))

    # Create lagged links (can be cyclic)
    lagged_links = []
    for l in range(L_lagged):

        cause = random_state.choice(causal_order)
        effect = random_state.choice(causal_order)
        while (cause, effect) in chosen_links or cause == effect:
            cause = random_state.choice(causal_order)
            effect = random_state.choice(causal_order)
        
        lagged_links.append((cause, effect))
        chosen_links.append((cause, effect))

    # print(chosen_links)
    # print(contemp_links)
    for (i, j) in chosen_links:

        # Choose lag
        if (i, j) in contemp_links:
            tau = 0
        else:
            tau = int(random_state.randint(1, tau_max+1))
        # print tau
        # CHoose coupling
        c = float(coupling_coeffs[random_state.randint(0, c_len)])
        if c != 0:
            func = coupling_funcs[random_state.randint(0, func_len)]

            links[j].append(((int(i), -tau), c, func))

    # # Stationarity check assuming model with linear dependencies at least for large x
    # # if check_stationarity(links)[0]:
    #     # return links
    # X, nonstat = toys.structural_causal_process(links, 
    #     T=100000, noises=None)
    # if nonstat == False:
    #     return links
    # else:
    #     print("Trial %d: Not a stationary model" % ir)


    # print("No stationary models found in {} trials".format(num_trials))
    # return None
    return links

@pytest.fixture(params=[
    # Generate a test data sample
    # Parameterize the sample by setting the autocorrelation value, coefficient
    # value, total time length, and random seed to different numbers
    # links_coeffs,               time,  seed_val
    (a_collider(0., 0.7),           1000,  2),
    (a_collider(0.6, 0.7),          1000,  2),
    (a_rule1(0., 0.7),          1000,  2),
    (a_rule2(0., 0.7),          1000,  2),
    (a_rule3(0., 0.7),          1000,  2),
    (a_rule3(0.5, 0.5),          1000,  2),

    ## Randomly generated structural causal processes, NEED to check whether seed
    ## generates graph that has only ONE member in Markov equivalence class!
    (a_random_process(
     N=5, L=5, coupling_coeffs=[0.7, -0.7],
     coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
     tau_max=5, contemp_fraction=0.3, num_trials=1,
     model_seed=3),             1000, 5),

    (a_random_process(
     N=5, L=5, coupling_coeffs=[0.7, -0.7],
     coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
     tau_max=5, contemp_fraction=0.3, num_trials=1,
     model_seed=4),             1000, 5),

    (a_random_process(
     N=5, L=5, coupling_coeffs=[0.7, -0.7],
     coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
     tau_max=5, contemp_fraction=0.3, num_trials=1,
     model_seed=5),             1000, 5),

    (a_random_process(
        N=10, L=10, coupling_coeffs=[0.7, -0.7],
        coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
        tau_max=5, contemp_fraction=0.3, num_trials=1,
        model_seed=3), 1000, 5),

    (a_random_process(
        N=20, L=20, coupling_coeffs=[0.7, -0.7],
        coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
        tau_max=5, contemp_fraction=0.2, num_trials=1,
        model_seed=4), 1000, 5),

    # (a_random_process(
    #     N=10, L=20, coupling_coeffs=[0.7, -0.7],
    #     coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
    #     tau_max=5, contemp_fraction=0.5, num_trials=1,
    #     model_seed=4), 1000, 6),

    # (a_random_process(
    #     N=8, L=20, coupling_coeffs=[0.7, -0.7],
    #     coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
    #     tau_max=5, contemp_fraction=0.5, num_trials=1,
    #     model_seed=4), 1000, 6),

    # (a_random_process(
    #     N=8, L=20, coupling_coeffs=[0.7, -0.7],
    #     coupling_funcs=[lin_f, lin_f], auto_coeffs=[0., 0.5],
    #     tau_max=5, contemp_fraction=0.5, num_trials=1,
    #     model_seed=5), 1000, 7),
])

def a_pcmciplus(request):
    # Set the parameters
    links_coeffs, time, seed_val = request.param

    # Retrieve lags
    tau_min, tau_max = toys._get_minmax_lag(links_coeffs)
    # Generate the data
    data, _ = toys.structural_causal_process(links=links_coeffs, T=time,
                                           noises=None, seed=seed_val)
    # Get the true parents
    # true_parents = toys._get_parents(links_coeffs, exclude_contemp=False)
    true_graph = toys.links_to_graph(links_coeffs, tau_max=tau_max)
    return pp.DataFrame(data), true_graph, links_coeffs, tau_min, tau_max

@pytest.fixture(params=[
    # Keep parameters for the algorithm here
    # pc_alpha, contemp_collider_rule, conflict_resolution, reset, cond_ind_test
    # OracleCI tests
    (0.01, 'majority', True, False, 'oracle_ci'),
    (0.01, 'none', True, False, 'oracle_ci'),
    (0.01, 'conservative', True, False, 'oracle_ci'),
    (0.01, 'majority', False, False, 'oracle_ci'),
    (0.01, 'none', False, False, 'oracle_ci'),
    (0.01, 'conservative', False, False, 'oracle_ci'),
    (0.01, 'majority', True, True, 'oracle_ci'),
    (0.01, 'none', True, True, 'oracle_ci'),
    (0.01, 'conservative', True, True, 'oracle_ci'),
    (0.01, 'majority', False, True, 'oracle_ci'),
    (0.01, 'none', False, True, 'oracle_ci'),
    (0.01, 'conservative', False, True, 'oracle_ci'),

    # ParCorr tests (can have finite sample errors)
    # (0.01, 'majority', True, False, 'par_corr'),
    # (0.01, 'none', True, False, 'par_corr'),
    # (0.01, 'conservative', True, False, 'par_corr'),
    # (0.01, 'majority', False, False, 'par_corr'),
    # (0.01, 'none', False, False, 'par_corr'),
    # (0.01, 'conservative', False, False, 'par_corr'),
    # (0.01, 'majority', True, True, 'par_corr'),
    # (0.01, 'none', True, True, 'par_corr'),
    # (0.01, 'conservative', True, True, 'par_corr'),
    # (0.01, 'majority', False, True, 'par_corr'),
    # (0.01, 'none', False, True, 'par_corr'),
    # (0.01, 'conservative', False, True, 'par_corr'),

])
def a_pcmciplus_params(request):
    # Return the parameters for the mci test
    return request.param

@pytest.fixture()
def a_run_pcmciplus(a_pcmciplus, a_pcmciplus_params):
    # Unpack the pcmci and the true parents, and common parameters
    dataframe, true_graph, links_coeffs, tau_min, tau_max = a_pcmciplus

    # Unpack the parameters
    (pc_alpha,
     contemp_collider_rule, conflict_resolution,
     reset_lagged_links, cond_ind_test_class,
     ) = a_pcmciplus_params

    if cond_ind_test_class == 'oracle_ci':
        cond_ind_test = OracleCI(links=links_coeffs)
    elif cond_ind_test_class == 'par_corr':
        cond_ind_test = ParCorr()

    # Run the PCMCI algorithm with the given parameters
    pcmci = PCMCI(dataframe=dataframe, cond_ind_test=cond_ind_test, verbosity=2)
    results = pcmci.run_pcmciplus(
                      link_assumptions=None,
                      tau_min=tau_min,
                      tau_max=tau_max,
                      pc_alpha=pc_alpha,
                      contemp_collider_rule=contemp_collider_rule,
                      conflict_resolution=conflict_resolution,
                      reset_lagged_links=reset_lagged_links,
                      max_conds_dim=None,
                      max_conds_py=None,
                      max_conds_px=None,
                      )
    # Print true links
    print("************************")
    print("\nTrue Graph")
    for lag in range(tau_max):
        print("Lag %d = ", lag)
        print(true_graph[:,:,lag])
    # pcmci.print_significant_links(
    #                             p_matrix=(true_graph != ""),
    #                             val_matrix=true_graph,
    #                             conf_matrix=None,
    #                             q_matrix=None,
    #                             graph=true_graph,
    #                             ambiguous_triples=None,
    #                             alpha_level=0.05)
    # Return the results and the expected result
    return results['graph'], true_graph

def test_pcmciplus(a_run_pcmciplus):
    """
    Test the pcmciplus algorithm and check it calculates the correct graph.
    """
    # Unpack the calculated and true graph
    graph, true_graph = a_run_pcmciplus
    # Ensure they are the same
    np.testing.assert_equal(graph, true_graph)


@pytest.fixture(params=[
    # Generate a test data sample
    # Parameterize the sample by setting the autocorrelation value, coefficient
    # value, total time length, and random seed to different numbers
    # links_coeffs,               time,  seed_val
    (a_random_process(
        N=5, L=15, coupling_coeffs=[0.5, -0.5],
        coupling_funcs=[lin_f], auto_coeffs=[0.5, 0.7],
        tau_max=1, contemp_fraction=0.6, num_trials=100,
        model_seed=2), 30, 5),

    (a_random_process(
        N=5, L=15, coupling_coeffs=[0.5, -0.5],
        coupling_funcs=[lin_f], auto_coeffs=[0.5, 0.7],
        tau_max=1, contemp_fraction=0.6, num_trials=100,
        model_seed=3), 30, 6),
])

def a_pcmciplus_order_independence(request):
    # Set the parameters
    links_coeffs, time, seed_val = request.param

    # Retrieve lags
    tau_min, tau_max = toys._get_minmax_lag(links_coeffs)
    # Generate the data
    data, _ = toys.structural_causal_process(links=links_coeffs, T=time,
                                           noises=None, seed=seed_val)
    # Get the true parents
    # true_parents = toys._get_parents(links_coeffs, exclude_contemp=False)
    true_graph = toys.links_to_graph(links_coeffs, tau_max=tau_max)
    return pp.DataFrame(data), true_graph, links_coeffs, tau_min, tau_max

@pytest.fixture(params=[
    # Keep parameters for the algorithm here
    # pc_alpha, contemp_collider_rule, conflict_resolution, reset, cond_ind_test
    # OracleCI tests

    # ParCorr tests (can have finite sample errors)
    (0.2, 'majority', True, False, 'par_corr'),
    (0.2, 'conservative', True, False, 'par_corr'),
    (0.2, 'majority', True, True, 'par_corr'),
    (0.2, 'conservative', True, True, 'par_corr'),
    (None, 'majority', True, False, 'par_corr'),
    (None, 'conservative', True, False, 'par_corr'),
    (None, 'majority', True, True, 'par_corr'),
    (None, 'conservative', True, True, 'par_corr'),
    ])
def a_pcmciplus_params_order_independence(request):
    # Return the parameters for the mci test
    return request.param

# @pytest.fixture()
def test_order_independence_pcmciplus(a_pcmciplus_order_independence,
                    a_pcmciplus_params_order_independence):
    # Unpack the pcmci and the true parents, and common parameters
    dataframe, true_graph, links_coeffs, tau_min, tau_max = \
        a_pcmciplus_order_independence

    data = dataframe.values[0]
    T = dataframe.T[0]
    N = dataframe.N

    # Unpack the parameters
    (pc_alpha,
     contemp_collider_rule, conflict_resolution,
     reset_lagged_links, cond_ind_test_class,
     ) = a_pcmciplus_params_order_independence

    if cond_ind_test_class == 'oracle_ci':
        cond_ind_test = OracleCI(links=links_coeffs)
    elif cond_ind_test_class == 'par_corr':
        cond_ind_test = ParCorr()

    # Run the PCMCI algorithm with the given parameters
    pcmci = PCMCI(dataframe=dataframe, cond_ind_test=cond_ind_test, verbosity=1)
    print("************************")
    print("\nTrue Graph")
    for lag in range(tau_max):
        print("Lag %d = ", lag)
        print(true_graph[:,:,lag])
    # pcmci.print_significant_links(
    #                             p_matrix=(true_graph == 0),
    #                             val_matrix=true_graph,
    #                             conf_matrix=None,
    #                             q_matrix=None,
    #                             graph=true_graph,
    #                             ambiguous_triples=None,
    #                             alpha_level=0.05)

    results = pcmci.run_pcmciplus(
                      link_assumptions=None,
                      tau_min=tau_min,
                      tau_max=tau_max,
                      pc_alpha=pc_alpha,
                      contemp_collider_rule=contemp_collider_rule,
                      conflict_resolution=conflict_resolution,
                      reset_lagged_links=reset_lagged_links,
                      max_conds_dim=None,
                      max_conds_py=None,
                      max_conds_px=None,
                      )
    correct_results = results['graph']

    for perm in itertools.permutations(range(N)):

        print(perm)
        data_new = np.copy(data[:,perm])
        dataframe = pp.DataFrame(data_new, var_names=list(perm))
        pcmci = PCMCI(dataframe=dataframe, cond_ind_test=cond_ind_test,
                      verbosity=1)
        results = pcmci.run_pcmciplus(
                      link_assumptions=None,
                      tau_min=tau_min,
                      tau_max=tau_max,
                      pc_alpha=pc_alpha,
                      contemp_collider_rule=contemp_collider_rule,
                      conflict_resolution=conflict_resolution,
                      reset_lagged_links=reset_lagged_links,
                      max_conds_dim=None,
                      max_conds_py=None,
                      max_conds_px=None,
                      )

        tmp = np.take(correct_results, perm, axis=0)
        back_converted_result = np.take(tmp, perm, axis=1)

        for tau in range(tau_max+1):
            if not np.array_equal(results['graph'][:,:,tau],
                               back_converted_result[:,:,tau]):
                print(tau)
                print(results['graph'][:,:,tau])
                print(back_converted_result[:,:,tau])
                print(back_converted_result[:,:,tau]-results['graph'][:,:,tau])
                print(perm)

        # np.allclose(results['graph'], back_converted_result)
        np.testing.assert_equal(results['graph'], back_converted_result)




"""Tigramite causal discovery for time series."""

# Author: Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

from __future__ import print_function
from scipy import stats, linalg
import numpy as np
import sys
import warnings

from tigramite.independence_tests.independence_tests_base import CondIndTest

class ParCorrMult(CondIndTest):
    r"""Partial correlation test for multivariate X and Y.

    Multivariate partial correlation is estimated through ordinary least squares (OLS)
    regression and some test for multivariate dependency among the residuals.

    Notes
    -----
    To test :math:`X \perp Y | Z`, first :math:`Z` is regressed out from
    :math:`X` and :math:`Y` assuming the  model

    .. math::  X & =  Z \beta_X + \epsilon_{X} \\
        Y & =  Z \beta_Y + \epsilon_{Y}

    using OLS regression. Then different measures for the dependency among the residuals 
    can be used. Currently only a test for zero correlation on the maximum of the residuals' 
    correlation is performed.

    Parameters
    ----------
    correlation_type : {'max_corr'}
        Which dependency measure to use on residuals.
    **kwargs :
        Arguments passed on to Parent class CondIndTest.
    """
    # documentation
    @property
    def measure(self):
        """
        Concrete property to return the measure of the independence test
        """
        return self._measure

    def __init__(self, correlation_type='max_corr', **kwargs):
        self._measure = 'par_corr_mult'
        self.two_sided = True
        self.residual_based = True

        self.correlation_type = correlation_type

        if self.correlation_type not in ['max_corr', 'PCCA_WilksLambda']:
            raise ValueError("correlation_type must be in ['max_corr', 'PCCA_WilksLambda']")

        CondIndTest.__init__(self, **kwargs)

    def _get_single_residuals(self, array, xyz, target_var,
                              standardize=True,
                              return_means=False):
        """Returns residuals of linear multiple regression.

        Performs a OLS regression of the variable indexed by target_var on the
        conditions Z. Here array is assumed to contain X and Y as the first two
        rows with the remaining rows (if present) containing the conditions Z.
        Optionally returns the estimated regression line.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        target_var : {0, 1}
            Variable to regress out conditions from.

        standardize : bool, optional (default: True)
            Whether to standardize the array beforehand. Must be used for
            partial correlation.

        return_means : bool, optional (default: False)
            Whether to return the estimated regression line.

        Returns
        -------
        resid [, mean] : array-like
            The residual of the regression and optionally the estimated line.
        """

        dim, T = array.shape
        dim_z = (xyz == 2).sum()

        # Standardize
        if standardize:
            array -= array.mean(axis=1).reshape(dim, 1)
            std = array.std(axis=1)
            for i in range(dim):
                if std[i] != 0.:
                    array[i] /= std[i]
            if np.any(std == 0.) and self.verbosity > 0:
                warnings.warn("Possibly constant array!")
            # array /= array.std(axis=1).reshape(dim, 1)
            # if np.isnan(array).sum() != 0:
            #     raise ValueError("nans after standardizing, "
            #                      "possibly constant array!")

        y = array[np.where(xyz==target_var)[0], :].T.copy()

        if dim_z > 0:
            z = (array[np.where(xyz==2)[0], :]).T.copy()
            beta_hat = np.linalg.lstsq(z, y, rcond=None)[0]
            mean = np.dot(z, beta_hat)
            resid = y - mean
        else:
            resid = y
            mean = None

        if return_means:
            return (np.fastCopyAndTranspose(resid), np.fastCopyAndTranspose(mean))

        return resid.T.copy()

    def get_dependence_measure(self, array, xyz, data_type=None):
        """Return multivariate kernel correlation coefficient.

        Estimated as some dependency measure on the
        residuals of a linear OLS regression.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        Returns
        -------
        val : float
            Partial correlation coefficient.
        """

        dim, T = array.shape
        dim_x = (xyz==0).sum()
        dim_y = (xyz==1).sum()

        x_vals = self._get_single_residuals(array, xyz, target_var=0)
        y_vals = self._get_single_residuals(array, xyz, target_var=1)

        array_resid = np.vstack((x_vals.reshape(dim_x, T), y_vals.reshape(dim_y, T)))
        xyz_resid = np.array([index_code for index_code in xyz if index_code != 2])

        val = self.mult_corr(array_resid, xyz_resid)

        return val

    def compute_wilks_lambda_cca(self, X, Y, num_components=None):
        """
        Compute Wilks Lambda from canonical correlations between X and Y using NumPy.
        
        Parameters:
        - X, Y: Arrays of shape (n_features, n_samples)
        - num_components: Number of canonical correlations to return (default: all)
        
        Returns:
        - wilks_lambda: Wilks Lambda of canonical correlations
        """

        # Center variables (per feature)
        X = X - X.mean(axis=1, keepdims=True)
        Y = Y - Y.mean(axis=1, keepdims=True)

        # Compute covariance matrices (features are rows, samples are columns)
        Cxx = np.cov(X, rowvar=True)
        Cyy = np.cov(Y, rowvar=True)
        Cxy = np.cov(X, Y, rowvar=True)[:X.shape[0], X.shape[0]:]

        # Ensure matrices are at least 2D
        if Cxx.ndim == 0:
            Cxx = np.array([[Cxx]])
        if Cyy.ndim == 0:
            Cyy = np.array([[Cyy]])
        if Cxy.ndim == 1:
            Cxy = Cxy[:, None]

        # Regularization for stability
        eps = 1e-10
        Cxx += eps * np.eye(Cxx.shape[0])
        Cyy += eps * np.eye(Cyy.shape[0])

        # Solve generalized eigenvalue problem
        M = linalg.inv(Cxx) @ Cxy @ linalg.inv(Cyy) @ Cxy.T

        eigvals = np.linalg.eigvalsh(M)
        eigvals = np.flip(np.real(eigvals))  # sort descending
        rho_sq = np.abs(np.maximum(eigvals, 0.0))

        if num_components is not None:
            rho_sq = rho_sq[:num_components]

        wilks_lambda = np.prod([1 - rho for rho in rho_sq])
        
        return wilks_lambda

    def mult_corr(self, array, xyz, standardize=True):
        """Return multivariate dependency measure.

        Parameters
        ----------
        array : array-like
            data array with X, Y in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        standardize : bool, optional (default: True)
            Whether to standardize the array beforehand. Must be used for
            partial correlation.

        Returns
        -------
        val : float
            Multivariate dependency measure.
        """

        dim, n = array.shape
        dim_x = (xyz==0).sum()
        dim_y = (xyz==1).sum()

        # Standardize
        if standardize:
            array -= array.mean(axis=1).reshape(dim, 1)
            std = array.std(axis=1)
            for i in range(dim):
                if std[i] != 0.:
                    array[i] /= std[i]
            if np.any(std == 0.) and self.verbosity > 0:
                warnings.warn("Possibly constant array!")
            # array /= array.std(axis=1).reshape(dim, 1)
            # if np.isnan(array).sum() != 0:
            #     raise ValueError("nans after standardizing, "
            #                      "possibly constant array!")

        x = array[np.where(xyz==0)[0]]
        y = array[np.where(xyz==1)[0]]

        if self.correlation_type == 'max_corr':
            # Get (positive or negative) absolute maximum correlation value
            corr = np.corrcoef(x, y)[:len(x), len(x):].flatten()
            val = corr[np.argmax(np.abs(corr))]

            # val = 0.
            # for x_vals in x:
            #     for y_vals in y:
            #         val_here, _ = stats.pearsonr(x_vals, y_vals)
            #         val = max(val, np.abs(val_here))
        elif self.correlation_type == 'PCCA_WilksLambda':
            val = self.compute_wilks_lambda_cca(x, y)

        # elif self.correlation_type == 'linear_hsci':
        #     # For linear kernel and standardized data (centered and divided by std)
        #     # biased V -statistic of HSIC reduces to sum of squared inner products
        #     # over all dimensions
        #     val = ((x.dot(y.T)/float(n))**2).sum()
        else:
            raise NotImplementedError("Currently only"
                                      "correlation_type == 'max_corr' and 'PCCA_WilksLambda' implemented.")

        return val

    def get_shuffle_significance(self, array, xyz, value,
                                 return_null_dist=False,
                                 data_type=None):
        """Returns p-value for shuffle significance test.

        For residual-based test statistics only the residuals are shuffled.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        value : number
            Value of test statistic for unshuffled estimate.

        Returns
        -------
        pval : float
            p-value
        """

        dim, T = array.shape
        dim_x = (xyz==0).sum()
        dim_y = (xyz==1).sum()

        x_vals = self._get_single_residuals(array, xyz, target_var=0)
        y_vals = self._get_single_residuals(array, xyz, target_var=1)

        array_resid = np.vstack((x_vals.reshape(dim_x, T), y_vals.reshape(dim_y, T)))
        xyz_resid = np.array([index_code for index_code in xyz if index_code != 2])


        null_dist = self._get_shuffle_dist(array_resid, xyz_resid,
                                           self.get_dependence_measure,
                                           sig_samples=self.sig_samples,
                                           sig_blocklength=self.sig_blocklength,
                                           verbosity=self.verbosity)

        # pval = (null_dist >= np.abs(value)).mean()
        pval = float(np.sum(null_dist >= value) + 1) / (self.sig_samples + 1)

        # Adjust p-value for two-sided measures
        if pval < 1.:
            pval *= 2.

        # Adjust p-value for dimensions of x and y (conservative Bonferroni-correction)
        # pval *= dim_x*dim_y

        if return_null_dist:
            return pval, null_dist
        return pval

    def get_analytic_significance(self, value, T, dim, xyz):
        """Returns analytic p-value depending on correlation_type.

        Assumes two-sided correlation. If the degrees of freedom are less than
        1, numpy.nan is returned.

        Parameters
        ----------
        value : float
            Test statistic value.

        T : int
            Sample length

        dim : int
            Dimensionality, ie, number of features.

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        Returns
        -------
        pval : float or numpy.nan
            P-value.
        """
        # Get the number of degrees of freedom
        deg_f = T - dim

        dim_x = (xyz==0).sum()
        dim_y = (xyz==1).sum()
        dim_z = dim - dim_x - dim_y

        if self.correlation_type == 'max_corr':
            if deg_f < 1:
                pval = np.nan
            elif abs(abs(value) - 1.0) <= sys.float_info.min:
                pval = 0.0
            else:
                trafo_val = value * np.sqrt(deg_f/(1. - value*value))
                # Two sided significance level
                pval = stats.t.sf(np.abs(trafo_val), deg_f) * 2
            # Adjust p-value for dimensions of x and y (conservative Bonferroni-correction)
            pval *= dim_x*dim_y
        elif self.correlation_type == 'PCCA_WilksLambda':
            if deg_f < 1:
                pval = np.nan
            else:
                trafo_val = -(T - dim_z - 0.5 * (dim_x + dim_y + 1)) * np.log(value)
                dof = dim_x * dim_y
                pval = 1 - stats.chi2.cdf(trafo_val, dof)     
        else:
            raise NotImplementedError("Currently only"
                                      "correlation_type == 'max_corr' and 'PCCA_WilksLambda' implemented.")
        

        return pval

    def get_model_selection_criterion(self, j, parents, tau_max=0, corrected_aic=False):
        """Returns Akaike's Information criterion modulo constants.

        Fits a linear model of the parents to each variable in j and returns
        the average score. Leave-one-out cross-validation is asymptotically
        equivalent to AIC for ordinary linear regression models. Here used to
        determine optimal hyperparameters in PCMCI, in particular the
        pc_alpha value.

        Parameters
        ----------
        j : int
            Index of target variable in data array.

        parents : list
            List of form [(0, -1), (3, -2), ...] containing parents.

        tau_max : int, optional (default: 0)
            Maximum time lag. This may be used to make sure that estimates for
            different lags in X, Z, all have the same sample size.

        Returns:
        score : float
            Model score.
        """

        Y = [(j, 0)]
        X = [(j, 0)]   # dummy variable here
        Z = parents
        array, xyz, _ = self.dataframe.construct_array(X=X, Y=Y, Z=Z,
                                                    tau_max=tau_max,
                                                    mask_type=self.mask_type,
                                                    return_cleaned_xyz=False,
                                                    do_checks=True,
                                                    verbosity=self.verbosity)

        dim, T = array.shape

        y = self._get_single_residuals(array, xyz, target_var=0)

        n_comps = y.shape[0]
        score = 0.
        for y_component in y:
            # Get RSS
            rss = (y_component**2).sum()
            # Number of parameters
            p = dim - 1
            # Get AIC
            if corrected_aic:
                comp_score = T * np.log(rss) + 2. * p + (2.*p**2 + 2.*p)/(T - p - 1)
            else:
                comp_score = T * np.log(rss) + 2. * p
            score += comp_score

        score /= float(n_comps)
        return score


if __name__ == '__main__':
    
    import tigramite
    from tigramite.data_processing import DataFrame
    # import numpy as np
    import timeit

    seed=None
    random_state = np.random.default_rng(seed=seed)
    cmi = ParCorrMult(
            correlation_type='PCCA_WilksLambda',
            # significance = 'shuffle_test',
            # sig_samples=1000,
        )

    samples=5000
    rate = np.zeros(samples)
    for i in range(samples):
        print(i)
        data = random_state.standard_normal((100, 6))
        # data[:,2] += -0.3*data[:,0]
        # data[:,1] += data[:,2]
        dataframe = DataFrame(data, 
            # vector_vars={0:[(0,0), (1,0)], 1:[(2,0),(3,0)], 2:[(4,0),(5,0)]}
            )

        cmi.set_dataframe(dataframe)

        pval = cmi.run_test(
                X=[(0,0), (1,0), ], 
                Y=[(2,0),(3, 0)], 
                # Z=[(5,0)]
                Z = [(4, 0), (5, 0)]
                )[1]
        
        rate[i] = pval <= 0.05

        # cmi.get_model_selection_criterion(j=0, parents=[(1, 0), (2, 0)], tau_max=0, corrected_aic=False)

        # print(cmi.run_test(X=[(0,0),(1,0)], Y=[(2,0), (3, 0)], Z=[(5,0)]))
    print(rate.mean())



"""Tigramite causal discovery for time series."""

# Author: Tom Hochsprung <tom.hochsprung@dlr.de>, Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

import numpy as np
import warnings

from scipy.stats import chi2, normaltest
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn import metrics
from sklearn.dummy import DummyClassifier

import tigramite
from tigramite.independence_tests.independence_tests_base import CondIndTest


class RegressionCI(CondIndTest):
    r"""Flexible parametric conditional independence tests for continuous, categorical, or mixed data.

    Asymptotically equivalent to the tests for mixed data suggested in

    Tsagris, Michail, et al. "Constraint-based causal discovery with mixed
    data." International journal of data science and analytics 6
    (2018): 19-30.

    For linear regression RegressionCI implements a likelihood ratio test,
    while the above employs a F-statistic. Furthermore, while our
    implementation utilizes the Chi^2 null distribution, theirs uses the
    F-distribution. 

    Assumes one-dimensional X, Y. But can be combined with PairwiseMultCI to
    obtain a test for multivariate X, Y.

    Notes
    -----
    To test :math:`X \perp Y | Z`, the regressions Y|XZ vs Y|Z, or, depending
    on certain criteria, X|YZ vs X|Z are compared. For that, the notion of
    the deviance is employed. If the fits of the respective regressions do
    not differ significantly (measured using the deviance), the null
    hypotheses of conditional independence is "accepted". This approach
    assumes that X and Y are univariate, and Z can be either empty,
    univariate or multivariate. Moreover, this approach works for all
    combinations of "discrete" and "continuous" X, Y and respective columns
    of Z; depending on the case, linear regression or multinomial regression
    is employed.

    Assumes one-dimensional X, Y.

    Parameters
    ----------
    **kwargs :
        Arguments passed on to parent class CondIndTest.
    """
    @property
    def measure(self):
        """
        Concrete property to return the measure of the independence test
        """
        return self._measure
    
    def __init__(self,
                 **kwargs):
        
        # Setup the member variables
        self._measure = 'regression_ci'
        self.two_sided = False
        self.residual_based = False
        self.recycle_residuals = False

        CondIndTest.__init__(self, **kwargs)

    def set_dataframe(self, dataframe):
        """Initialize and check the dataframe.

        Parameters
        ----------
        dataframe : data object
            Set tigramite dataframe object. It must have the attributes
            dataframe.values yielding a numpy array of shape (observations T,
            variables N) and optionally a mask of the same shape and a missing
            values flag.

        """
        self.dataframe = dataframe
        
        if self.mask_type is not None:
            if dataframe.mask is None:
                raise ValueError("mask_type is not None, but no mask in dataframe.")
            dataframe._check_mask(dataframe.mask)
        
        if dataframe.data_type is None:
            raise ValueError("data_type cannot be None for RegressionCI.")
        dataframe._check_mask(dataframe.data_type)

    def get_dependence_measure(self, array, xyz, data_type):
        """Returns test statistic.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns.

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        data_type : array-like
            array of same shape as array which describes whether samples
            are continuous or discrete: 0s for continuous and
            1s for discrete

        Returns
        -------
        val : float
            test estimate.
        """

        def convert_to_one_hot(data, nb_classes):
            """Convert an iterable of indices to one-hot encoded labels."""
            targets = np.array(data).reshape(-1).tolist()
            # categories need to be mapped to 0, 1, ... in this function
            s = sorted(set(targets))
            targets = [s.index(i) for i in targets]
            targets = np.array(targets)
            return np.eye(nb_classes)[targets]

        def do_componentwise_one_hot_encoding(X, var_type):
            """A function that one-hot encodes all categorical components of X"""
           
            T, dim = X.shape
            X_new = np.empty([T, 0])
            # componentwise dummy-encoding (if necessary, otherwise, keep component as usual):
            for i in range(0, len(var_type)):
                if var_type[i] == 1:
                    nb_classes = len(set(X[:, i]))
                    X_new = np.hstack((X_new, convert_to_one_hot(X[:, i].astype(int), nb_classes=nb_classes)))
                elif var_type[i] == 0:
                    X_new = np.hstack((X_new, X[:, i].reshape((T, 1))))
                else:
                    raise ValueError("data_type only allows entries in {0, 1}")
            return X_new

        def calc_deviance_logistic(X, y, var_type):
            """Calculates the deviance (i.e., 2 * log-likelihood) for a multinomial logistic regression
            (with standard regression assumptions)
            """

            # 1-hot-encode all categorical columns
            X = do_componentwise_one_hot_encoding(X, var_type=var_type)
            y = np.ravel(y).astype('int')
            # do logistic regression, if y only contains one class, return zero.
            if len(np.unique(y)) < 2:
                model = DummyClassifier(strategy="constant", constant=y[0])
                model.fit(X, y)
                deviance = 0.
                warnings.warn("Constant array detected, CI test ill-defined!")
            else:
                model = LogisticRegression(solver='lbfgs')
                model.fit(X, y)
                deviance = 2 * metrics.log_loss(y, model.predict_proba(X), normalize=False)
            
            # dofs: +2 for intercept (+1) (not too important, cancels out later anyway)
            dof = model.n_features_in_ + 1
            
            return deviance, dof

        def calc_deviance_linear(X, y, var_type):
            """Calculates the deviance (i.e., 2 * log-likelihood) for a linear regression
            (with standard regression assumptions
            """

            n, p = X.shape  # p is not important for later
            # 1-hot-encode all categorical columns
            X = do_componentwise_one_hot_encoding(X, var_type = var_type)
            y = np.ravel(y)
            if y.std() == 0.: 
                warnings.warn("Constant array detected, CI test ill-defined!")
            # do linear regression
            model = LinearRegression()
            model.fit(X, y)
            # predictions based on fitted model
            preds = model.predict(X)
            # residual sum of squares
            rss = np.sum(np.power((preds - y), 2))
            # deviance (only the term with the rss-term is important, the rest cancels out later anyway)
            # deviance is calculated as -2*log-likelihood
            deviance = n * np.log(2 * np.pi) + n * np.log(rss / n) + n
            # dofs: +2 for intercept (+1)  (not too important, cancels out later anyway)
            dof = model.n_features_in_  + 1
            return deviance, dof

        def entropy(series):
            value, counts = np.unique(series, return_counts=True)
            norm_counts = counts / counts.sum()
            return -(norm_counts * np.log(norm_counts)).sum()

        x_indices = np.where(xyz == 0)[0]
        y_indices = np.where(xyz == 1)[0]
        z_indices = np.where(xyz == 2)[0]

        x = array[x_indices].T
        y = array[y_indices].T

        x_type = data_type[x_indices]
        y_type = data_type[y_indices]

        if len(z_indices) == 0:
            z = np.ones((array.shape[1], 1))
            z_type = [0]
        else:
            z = array[z_indices].T
            z_type = data_type[z_indices]
            z_type = z_type.max(axis=1)

        # check, whether within X and within Y all datapoints have the same datatype
        if ((x_type.max() != x_type.min()) or (y_type.max() != y_type.min())):
            raise ValueError("All samples regarding X or respectively Y must have the same datatype")
        
        x_type = x_type.max()
        y_type = y_type.max()

        # if z was (originally) None, then just an intercept is fitted ...
        # Now, different cases for X discrete/continuous and Y discrete/continuous
        
        # Case 1: X continuous, Y continuous
        if (x_type == 0) and (y_type == 0):
            # Use the more normal variable as dependent variable TODO: makes sense?
            if normaltest(x)[0] >= normaltest(y)[0]:
                dep_var = y
                rest = np.hstack((x, z))
                rest_type = np.hstack((x_type, z_type))
            else:
                dep_var = x
                rest = np.hstack((y, z))  
                rest_type = np.hstack((y_type, z_type))
              
            # Fit Y | Z
            dev1, dof1 = calc_deviance_linear(z, dep_var, var_type = z_type)
            # Fit Y | ZX
            dev2, dof2 = calc_deviance_linear(rest, dep_var, var_type=rest_type)
            # print(dev1, dev2, np.abs(dev1 - dev2))

        # Case 2: X discrete, Y continuous
        elif (x_type == 1) and (y_type == 0):
            xz = np.hstack((x, z))
            # Fit Y | Z
            dev1, dof1 = calc_deviance_linear(z, y, var_type = z_type)
            # Fit Y | XZ
            dev2, dof2 = calc_deviance_linear(xz, y, var_type = np.hstack((x_type, z_type)))
        
        # Case 3: X continuous, Y discrete
        elif (x_type == 0) and (y_type == 1):
            yz = np.hstack((y, z))
            # Fit X | Z
            dev1, dof1 = calc_deviance_linear(z, x, var_type = z_type)
            # Fit X | YZ
            dev2, dof2 = calc_deviance_linear(yz, x, var_type = np.hstack((y_type, z_type)))
        
        # Case 4: X discrete, Y discrete
        elif (x_type == 1) and (y_type == 1):
            # Use the variable with smaller entropy as dependent variable TODO: makes sense?
            if entropy(x) >= entropy(y):
                dep_var = y
                rest = np.hstack((x, z))
                rest_type = np.hstack((x_type, z_type))
            else:
                dep_var = x
                rest = np.hstack((y, z))  
                rest_type = np.hstack((y_type, z_type))
            # xz = np.hstack((x, z))
            # Fit Y | Z
            dev1, dof1 = calc_deviance_logistic(z, dep_var, var_type = z_type)
            # Fit Y | XZ
            dev2, dof2 = calc_deviance_logistic(rest, dep_var, var_type=rest_type)

        # calculate the difference between the deviance for the smaller and for the larger model
        # (i.e., the actual deviance)
        stat = dev1 - dev2
        dof = dof2 - dof1

        self._temp_dof = dof
        return stat

    def get_analytic_significance(self, value, T, dim, xyz):
        """Return the p_value of test statistic.

        According to a chi-square distribution with 'dof' degrees of freedom.

        """
                      
        # Calculate the p_value
        p_value = chi2.sf(value, self._temp_dof)
        del self._temp_dof

        return p_value


if __name__ == '__main__':
    

    import pandas as pd
    import numpy as np
    import matplotlib.pyplot as plt
    from tigramite.pcmci import PCMCI
    # from tigramite.independence_tests.regressionCI import RegressionCI
    import tigramite.plotting as tp
    import tigramite.data_processing as pp

    import tigramite
    from tigramite.data_processing import DataFrame
    import tigramite.data_processing as pp
    import numpy as np

    seed=43
    random_state = np.random.default_rng(seed=seed)
    ci = RegressionCI()

    T = 100

    reals = 100
    rate = np.zeros(reals)

    x_example = "continuous"
    y_example = "continuous"
    dimz = 1
    # z_example = ["discrete", "continuous"]
    z_example = ["continuous"] #, "discrete"]
    # z_example = None
    rate = np.zeros(reals)
    for i in range(reals):
        if (dimz > 0):
            z = np.zeros((T, dimz))
            for k in range(0, len(z_example)):
                if z_example[k] == "discrete":
                    z[:, k] = random_state.binomial(n=1, p=0.5, size=T)
                else:
                    z[:, k] = random_state.uniform(low = 0, high = 1, size=T)
        else:
            z = None
        x = np.empty(T).reshape(T, 1)
        y = np.empty(T).reshape(T, 1)
        for t in range(T):
            if dimz > 0:
                if z_example[0] == "discrete":
                    val = z[t, 0].squeeze()
                    prob = 0.2 + val * 0.6
                else:
                    prob = z[t, 0].squeeze()
            else:
                prob = 0.2
            if x_example == "discrete":
                x[t] = random_state.choice([0, 1], p=[prob, 1. - prob])
            else:
                x[t] = 0.1*random_state.random() # np.random.uniform(prob, 1)  #np.random.normal(prob, 1)
            if y_example == "discrete":
                y[t] = random_state.choice([0, 1], p=[prob, (1. - prob)]) # + x[t]
            else:
                y[t] = random_state.normal(prob, 1) + 0.5*x[t]

        # # Continuous data
        # z = np.random.randn(T, dimz)
        # x = (0.5*z[:,0] + np.random.randn(T)).reshape(T, 1)
        # y = (0.5*z[:,0] + np.random.randn(T)).reshape(T, 1) #+ 2*x

        if x_example == "discrete":
            x_type = np.ones(T)
        else:
            x_type = np.zeros(T)
        if y_example == "discrete":
            y_type = np.ones(T)
        else:
            y_type = np.zeros(T)
        if dimz > 0:
            z_type = np.zeros((T, dimz))
            for j in range(0, len(z_example)):
                if z_example[j] == "discrete":
                    z_type[:, j] = np.ones(T)
                else:
                    z_type[:, j] = np.zeros(T)
        else:
            z_type = None

        val, pval = ci.run_test_raw(x, y, z=z, x_type=x_type, y_type=y_type, z_type=z_type)
        rate[i] = pval

        # data = np.hstack((x, y, z))
        # data_type = np.zeros(data.shape)
        # data_type[:, 0] = x_example == "discrete"
        # data_type[:, 1] = y_example == "discrete"
        # data_type[:, 2] = z_example == "discrete"
        # data_type = data_type.astype('int')
        # # print(data_type)
        # dataframe = pp.DataFrame(data=data, data_type=data_type)
        # ci.set_dataframe(dataframe)
        
        # val, pval = ci.run_test(X=[(0, 0)], Y=[(1, 0)], Z=[(2, 0)])
        # rate[i] = pval

    print((rate <= 0.05).mean())


# dummy



"""Tigramite causal discovery for time series."""

# Authors: Elena Saggioro, Sagar Simha, Matthias Bruhns, Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

from copy import deepcopy
import numpy as np
import sklearn
from joblib import Parallel, delayed
from ortools.linear_solver import pywraplp
import traceback

from tigramite.independence_tests.parcorr import ParCorr
from tigramite.data_processing import DataFrame
from tigramite.models import Prediction
from tigramite.pcmci import PCMCI

class RPCMCI(PCMCI):
    r"""RPCMCI class for extracting causal regimes and the associated graphs from
        time series data.

        Notes
        -----
        The Regime-PCMCI causal discovery method is described in: 

        Elena Saggioro, Jana de Wiljes, Marlene Kretschmer, Jakob Runge;
        Reconstructing regime-dependent causal relationships from observational
        time series. Chaos 1 November 2020; 30 (11): 113115.
        https://doi.org/10.1063/5.0020538

        The method iterates between two phases --a regime learning phase
        (optimization-based) and a causal discovery phase (PCMCI)-- to identify
        regime dependent causal relationships. A persistent discrete regime
        variable is assumed that leads to a finite number of regimes within which
        stationarity can be assumed.

        Parameters
        ----------
        dataframe : data object
            This is the Tigramite dataframe object. It has the attributes
            dataframe.values yielding a numpy array of shape ( observations T,
            variables N). For RPCMCI the mask will be ignored. You may use the
            missing_flag to indicate missing values.
        cond_ind_test : conditional independence test object
            This can be ParCorr or other classes from
            ``tigramite.independence_tests`` or an external test passed as a
            callable. This test can be based on the class
            tigramite.independence_tests.CondIndTest.
        prediction_model : sklearn model object
            For example, sklearn.linear_model.LinearRegression() for a linear
            regression model. This should be consistent with cond_ind_test, ie, 
            use ParCorr() with a linear model and, eg, GPDC() with a 
            GaussianProcessRegressor model, or CMIknn with NearestNeighbors model.
        seed : int
            Random seed for annealing step.
        verbosity : int, optional (default: -1)
            Verbose levels -1, 0, 1, ...
        """

    def __init__(self, dataframe, cond_ind_test=None, 
                    prediction_model=None, seed=None, verbosity=-1):

        self.verbosity = verbosity

        self.seed = seed
        if self.seed is None:
            self.seed = np.random.randint(0, 1000)

        # Set prediction model to be used in optimization
        self.prediction_model = prediction_model
        if self.prediction_model is None:
            self.prediction_model = sklearn.linear_model.LinearRegression()

        # Set conditional independence test
        if cond_ind_test is None:
            cond_ind_test = ParCorr()
        cond_ind_test.set_mask_type('y')

        if dataframe.analysis_mode != 'single':
            raise ValueError("Only single time series data allowed for RPCMCI.")
   
        if dataframe.has_vector_data:
            raise ValueError("Only scalar data allowed for RPCMCI.")
        
               
        # Masking is not available in RPCMCI, but missing values can be specified
        dataframe.mask = {0:np.zeros(dataframe.values[0].shape, dtype='bool')}
        self.missing_flag = dataframe.missing_flag

        # Init base class
        PCMCI.__init__(self, dataframe=dataframe,
                        cond_ind_test=cond_ind_test,
                        verbosity=0)

    def run_rpcmci(self,
                   num_regimes,
                   max_transitions,
                   switch_thres=0.05,
                   num_iterations=20,
                   max_anneal=10,
                   tau_min=1,
                   tau_max=1,
                   pc_alpha=0.2,
                   alpha_level=0.01,
                   n_jobs=-1,
                   ):

        """Run RPCMCI method for extracting causal regimes and the associated graphs from
            time series data.

        Parameters
        ----------
        num_regimes : int
            Number of assumed regimes.
        max_transitions : int
            Maximum number of transitions within a single regime (persistency parameter).
        switch_thres : float
            Switch threshold.
        num_iterations : int
            Optimization iterations.
        max_anneal : int
            Maximum annealing runs.
        tau_min : int, optional (default: 0)
            Minimum time lag to test.
        tau_max : int, optional (default: 1)
            Maximum time lag. Must be larger or equal to tau_min.
        pc_alpha : float, optional (default: 0.2)
            Significance level in PCMCI.
        alpha_level : float, optional (default: 0.05)
            Significance level in PCMCI at which the p_matrix is thresholded to
            get graph.
        n_jobs : int, optional (default: -1)
            Number of CPUs to use in joblib parallization. Default n_jobs=-1
            uses all available.

        Returns
        -------
        regimes : array of shape (n_regimes, T)
            One-hot encoded regime variable.
        causal_results: dictionary
            Contains result of run_pcmci() after convergence.
        diff_g_f : tuple
            Difference between two consecutive optimizations for all annealings and
            the optimal one with minimum objective value (see paper).
        error_free_annealings : int
            Number of annealings that converged without error.
        """

        count_saved_ann = 0
        # initialize residuals (objective value) of MIP optimize
        objmip_ann = [None] * max_anneal
        parents_ann = [None] * max_anneal
        causal_prediction = [None] * max_anneal
        links_ann = [None] * max_anneal
        gamma_ann = [None] * max_anneal
        diff_g_ann = [None] * max_anneal
        q_break_cycle = 5

        data = self.dataframe.values[0]

        def _pcmci(tau_min, tau_max, pc_alpha, alpha_level):
            """Wrapper around running PCMCI."""
            results = self.run_pcmci(tau_min=tau_min, tau_max=tau_max, pc_alpha=pc_alpha, alpha_level=alpha_level)
            graph = results['graph'] 
            pcmci_parents = self.return_parents_dict(graph=graph, val_matrix=results['val_matrix'])
            return results, graph, pcmci_parents

        def _optimize_gamma(resid_sq, max_transitions):
            r"""
            Solves the following optimization problem :

                minimize       c * x

            where c = resid_sq , flattened along num_regimes dimension
                  x = Gamma , flattened along num_regimes dimension

            with Constraints:
                    (1) [\sum_{k=1,num_regimes}gamma^k(t) ]= 1
                                            forall t    : uniqueness
                    (2) [\sum_{t=1:T-1} | gamma^k(t+1) - gamma^k(t) | ] <= max_transitions
                                            forall k    : persistence


            Inputs:
             resid_sq ( np.shape = (num_regimes,T) )
             max_transitions = max number of switchings allowed

            Returns:
            Gamma_updated ( np.shape = (num_regimes,T) ))
            """

            num_regimes, T = resid_sq.shape

            # Create the linear solver with the GLOP backend.
            solver = pywraplp.Solver.CreateSolver("GLOP")
            infinity = solver.infinity()

            # Define vector of integer variables in the interval [0,1].
            G = [solver.NumVar(0, 1, f"x_{i}") for i in range(num_regimes * T)]

            # Define eta, auxiliary vars for constr. (2).
            E = [solver.NumVar(0, infinity, f"eta_{i}") for i in range(num_regimes * T - 1)]
            X = G + E
            solver.Minimize(
                sum([resid_sq[k, t] * X[k * T + t] for k in range(num_regimes) for t in range(T)])
            )

            con_lst = [sum([X[k * T + t] for k in range(num_regimes)]) for t in range(T)]
            for t in range(T):
                solver.Add(con_lst[t] == 1)

            for k in range(num_regimes):
                for t in range(T - 1):
                    # (2.1)
                    solver.Add(
                        (X[k * T + t + 1] - X[k * T + t] - X[k * T + t + num_regimes * T] <= 0)
                    )
                    # (2.2)
                    solver.Add(
                        (
                            (
                                    -1 * X[k * T + t + 1]
                                    + X[k * T + t]
                                    - X[k * T + t + num_regimes * T]
                                    <= 0
                            )
                        )
                    )
                    # (2.3)
                    solver.Add(
                        ((sum([X[k * T + t + num_regimes * T] for t in range(T - 1)]) <= max_transitions))
                    )

            status = solver.Solve()
            if status == pywraplp.Solver.OPTIMAL:
                if self.verbosity > -1:
                    print("\nOptimal objective: reached.")
                gamma = np.reshape([g.solution_value() for g in G], (num_regimes, T))
                obj_value = solver.Objective().Value()
                return gamma, obj_value
            else:
                if self.verbosity > -1:
                    print("The problem does not have an optimal solution. Please change hyperparameters.")
                exit(0)

        def one_annealing_step(a):
            """Executes one annealing step. The random seed is self.seed + a."""

            if self.verbosity > -1:
                print(f"\n################# Annealing iteration a = {a} ####################\n")

            T = self.dataframe.T[0]

            # Initialise gamma_0 as random matrix of 1s and 0s
            random_state = np.random.default_rng(self.seed + a)
            gamma_opt = random_state.uniform(0, 1, size=(num_regimes, T))  # range is [0,1)!

            parents_opt = {} # [None] * num_regimes
            results_opt = {} # [None] * num_regimes
            links_opt = {} # [None] * num_regimes
            objective_opt = 0

            # Difference between two consecutive optimizations
            diff_g = []

            #
            # Iteration over 1. causal discovery and 2. constrained optimization
            #
            error_flag = False
            for q in range(num_iterations):
                if self.verbosity > -1:
                    print(f"\n###### Optimization step q = {q}")

                # Initialize to 0
                residuals = np.zeros((num_regimes, T, self.N))

                gamma_temp = deepcopy(gamma_opt)

                #
                # 1. Causal discovery and prediction
                # 

                # Iterate over regimes
                for k in range(num_regimes):
                    if self.verbosity > -1:
                        print(f"{16 * '#'} Regime k = {k}")

                    # Select sample according to gamma_opt, is a bool vector
                    selected_samples_k = (gamma_temp[k, :] > switch_thres)

                    mask_of_k = np.ones(data.shape, dtype="bool")
                    mask_of_k[selected_samples_k] = False

                    # df_of_k = pp.DataFrame(data, mask=mask_of_k, missing_flag=self.missing_flag,
                    #              var_names=self.var_names)

                    # Change mask in dataframe for this step
                    self.dataframe.mask[0] = mask_of_k

                    if np.any((mask_of_k == False).sum(axis=0) <= 5):
                        error_flag = True
                        if self.verbosity > -1:
                            print(f"*****Regime with too few samples in annealing a = {a} at iteration q = {q}.*****\n")
                        if self.verbosity > -1:
                            print("***** Break k-loop of regimes *****\n ")
                        break  # from k-loop         

                    try: 
                        # cond_ind_test = getattr(self, method)(**method_args)
                        # pcmci = PCMCI(dataframe=df_of_k, 
                        #     cond_ind_test=self.cond_ind_test, 
                        #     verbosity=0)
                        results_temp, link_temp, parents_temp = _pcmci(
                                        # pcmci,
                                                                        tau_max=int(tau_max),
                                                                        pc_alpha=pc_alpha,
                                                                        alpha_level=alpha_level,
                                                                        tau_min=tau_min,)
                    except Exception:
                        traceback.print_exc()
                        error_flag = True
                        print(f"*****Value error in causal discovery for annealing a = {a} at iteration q = {q}.*****\n")
                        print("***** Break k-loop of regimes *****\n ")
                        break  # from k-loop

                    parents_opt[k] = parents_temp
                    results_opt[k] = results_temp
                    links_opt[k] = link_temp

                    try: 
                        # Prediction with causal parents
                        pred = Prediction(
                            dataframe=self.dataframe,
                            prediction_model=self.prediction_model,
                            data_transform=sklearn.preprocessing.StandardScaler(),
                            train_indices=range(T),
                            test_indices=range(T),
                            verbosity=0,
                        )

                        pred.fit(
                            target_predictors=parents_temp,
                            selected_targets=range(self.N),
                            tau_max=int(tau_max),
                        ) 
                        # print(parents_temp)
                        # Compute the predicted residuals for each variable
                        predicted = pred.predict(
                            target=list(range(self.N)), 
                            new_data=DataFrame(data, missing_flag=self.missing_flag)
                        )

                        original_data = np.zeros(predicted.shape)
                        for target in range(self.N):
                            # print(data.shape, predicted.shape, original_data.shape, pred.get_test_array(target).shape, mask_of_k.sum(axis=0))
                            # print(pred.get_test_array(target)[0].flatten().std())
                            original_data[:, target] = pred.get_test_array(target)[0].flatten()

                    except Exception:
                        traceback.print_exc()
                        error_flag = True
                        print(f"*****Value error in prediction for annealing a = {a} at iteration q = {q}.*****\n")
                        print("***** Break k-loop of regimes *****\n ")
                        break  # from k-loop


                    # Get residuals
                    residuals[k, int(tau_max):, :] = original_data - predicted
                    # print(np.abs(residuals[k, int(tau_max):, :]).mean(axis=0))

                if error_flag:
                    if self.verbosity > -1:
                        print(f"***** Break q-loop of optimization iterations for Annealing a = {a} at iteration q = {q}." 
                            " Go to next annealing step. *****\n")
                    break

                #
                # 2. Regime optimization step with side constraints
                #

                # Comute the resid_sq
                res_sq = np.square(residuals).sum(axis=-1)
                # print(res_sq.shape)

                try:
                    # Optimization
                    gamma_opt, objective_opt = _optimize_gamma(res_sq, max_transitions)

                except Exception:
                    traceback.print_exc()
                    error_flag = True
                    print(f"*****Value error in optimization for annealing a = {a} at iteration q = {q}.*****\n")
                    break  

                diff_g.append(np.sum(np.abs(gamma_opt - gamma_temp)))

                if self.verbosity > -1:
                    print(f"Difference in abs value between the previous and current gamma "
                        f"(shape num_regimesxT) : {diff_g[q]}")

                # Break conditions
                if diff_g[-1] == 0:
                    if self.verbosity > -1:
                        print("Two consecutive gammas are equal: (local) minimum reached. "
                            "Go to next annealing.\n")
                    break

                if (q >= q_break_cycle) and (diff_g[-1] <= (2 * num_regimes * T // 100)):
                    if self.verbosity > -1:
                        print(f"Iteration larger than {q_break_cycle} and two consecutive gammas are too similar. "
                        f"Go to next annealing.\n")
                    break

            if error_flag:
                if self.verbosity > -1:
                    print(f"*****Annealing a = {a} failed****\n")

                return None

            return a, objective_opt, parents_opt, results_opt, links_opt, gamma_opt, diff_g

        # Parallelizing over annealing steps
        all_results = Parallel(n_jobs=n_jobs)(
            delayed(one_annealing_step)(a) for a in range(max_anneal))

        # all_results = []
        # for a in range(max_anneal):
        #     all_results.append(one_annealing_step(a))

        error_free_annealings = 0
        for result in all_results:
            if result is not None:
                error_free_annealings += 1
                a, objective_opt, parents_opt, results_opt, links_opt, gamma_opt, diff_g = result
                
                # Save annealing results
                objmip_ann[a] = objective_opt  
                parents_ann[a] = parents_opt  
                causal_prediction[a] = results_opt
                links_ann[a] = links_opt
                gamma_ann[a] = gamma_opt 
                diff_g_ann[a] = diff_g 

        if error_free_annealings == 0:
            print("No annealings have converged. Run failed.")
            return None

        # If annealing values are larger than the default. 
        # Can happen for long time series and high dimensionality
        min_obj_val = np.min([a for a in objmip_ann if a is not None])
        i_best = objmip_ann.index(min_obj_val)

        # Final results based on best
        # parents_f = parents_ann[i_best]
        results_f = causal_prediction[i_best]
        # links_f = links_ann[i_best]
        gamma_f = gamma_ann[i_best]
        # Convergence optimization
        diff_g_f = diff_g_ann, diff_g_ann[i_best]

        final_results = {'regimes': gamma_f,
                         'causal_results':results_f,
                         'diff_g_f':diff_g_f,
                         'error_free_annealings':error_free_annealings}
        
        return final_results


"""
Tests for independence_tests.py.
"""
from __future__ import print_function
from collections import OrderedDict
import numpy as np
import pytest
import sys

import tigramite.data_processing as pp
from tigramite.toymodels import structural_causal_processes as toys

# Pylint settings
# pylint: disable=redefined-outer-name

# Define the verbosity at the global scope
VERBOSITY = 1

# CONSTRUCT ARRAY TESTING FUNCTIONS ############################################
def rand_node(t_min, n_max, t_max=0, n_min=0):
    """
    Generate a random node to test
    """
    rand_node = np.random.randint(n_min, n_max)
    rand_time = np.random.randint(t_min, t_max)
    return (rand_node, rand_time)

def gen_nodes(n_nodes, seed, t_min, n_max):
    """
    Generate some random nodes to tests
    """
    # Set the seed if needed
    np.random.seed(seed)
    # Y nodes are always at (0, 0)
    y_nds = [(0, 0)]
    # X nodes is only one node
    x_nds = [rand_node(t_min, n_max)]
    # Z nodes are multiple nodes
    z_nds = [rand_node(t_min, n_max) for _ in range(n_nodes)]
    return x_nds, y_nds, z_nds


# CONSTRUCT ARRAY TESTING ######################################################
@pytest.fixture(params=[
    # Parameterize the array construction
    #(X, Y, Z) nodes,        t_max, m_val, mask_type
    (gen_nodes(3, 0, -3, 9), 3,     False, None),           # Few nodes
    (gen_nodes(7, 1, -3, 9), 3,     False, None),           # More nodes
    (gen_nodes(7, 2, -3, 3), 3,     False, None),           # Repeated nodes
    (gen_nodes(7, 3, -3, 9), 3,     True,  None),           # Missing vals
    (gen_nodes(7, 4, -3, 9), 3,     True,  ['x']),          # M-val + masked x
    (gen_nodes(7, 4, -3, 9), 3,     True,  ['x','y']),      # M-val + masked xy
    (gen_nodes(3, 5, -4, 9), 2,     False, ['x']),          # masked x
    (gen_nodes(3, 6, -4, 9), 2,     False, ['y']),          # masked y
    (gen_nodes(3, 7, -4, 9), 2,     False, ['z']),          # masked z
    (gen_nodes(3, 7, -4, 9), 2,     False, ['x','y','z'])]) # mask xyz
def cstrct_array_params(request):
    return request.param

def test_construct_array(cstrct_array_params):
    # Unpack the parameters
    (x_nds, y_nds, z_nds), tau_max, missing_vals, mask_type =\
        cstrct_array_params
    # Make some fake data
    data = np.arange(1000).reshape(10, 100).T.astype('float')
    # Get the needed parameters from the data
    T, N = data.shape
    max_lag = 2*tau_max
    n_times = T - max_lag

    # When testing masking and missing value flags, we will remove time slices,
    # starting with the earliest slice.  This counter keeps track of how many
    # rows have been masked.
    n_rows_masked = 0

    # Make a fake mask
    data_mask = np.zeros_like(data, dtype='bool')
    if mask_type is not None:
        for var, nodes in zip(['x', 'y', 'z'], [x_nds, y_nds, z_nds]):
            if var in mask_type:
                # Get the first node
                a_nd, a_tau = nodes[0]
                # Mask the first value of this node
                data_mask[a_tau - n_times + n_rows_masked, a_nd] = True
                n_rows_masked += 1

    # Choose fake missing value as the earliest time entry in the first z-node
    # from the original (non-shifted) datathat is not cutoff by max_lag or
    # masked values from the first z-node
    missing_flag = None
    if missing_vals:
        # Get the node index
        a_nd, _ = z_nds[0]
        # Select the earliest non-cutoff entry from the unshifted data set
        earliest_time = max_lag + n_rows_masked
        missing_flag = data[earliest_time, a_nd]
        # Record that the row with this value and all rows up to max_lag after
        # this value have been cut off as well
        # n_rows_masked += 1

    # Construct the array
    data_f = pp.DataFrame(data, data_mask, missing_flag)
    array, xyz, _ = data_f.construct_array(x_nds, y_nds, z_nds,
                                        tau_max=tau_max,
                                        mask_type=mask_type,
                                        verbosity=VERBOSITY)
    # Ensure x_nds, y_nds, z_ndes are unique
    x_nds = list(OrderedDict.fromkeys(x_nds))
    y_nds = list(OrderedDict.fromkeys(y_nds))
    z_nds = list(OrderedDict.fromkeys(z_nds))
    z_nds = [node for node in z_nds
             if (node not in x_nds) and (node not in y_nds)]

    # Get the expected results
    expect_array = np.array([list(np.arange(data[time-n_times, node],
                                        data[time-n_times, node]+n_times))
                             for node, time in x_nds + y_nds + z_nds]).astype('float')
    expect_xyz = np.array([0 for _ in x_nds] +\
                          [1 for _ in y_nds] +\
                          [2 for _ in z_nds])
    # Apply the mask, which always blocks the latest time of the 0th node of the
    # masked variable, which removes the first n time slices in the returned
    # array
    expect_array = expect_array[:, n_rows_masked:]
    if missing_vals:
        missing_anywhere_base = np.array(np.where(np.any(expect_array==missing_flag, axis=0))[0])
        missing_anywhere = list(missing_anywhere_base)
        # for tau in range(1, max_lag+1):
        #     missing_anywhere += list(np.array(missing_anywhere_base) + tau)
        expect_array = np.delete(expect_array, missing_anywhere, axis=1)

    # Test the results
    np.testing.assert_almost_equal(array, expect_array)
    np.testing.assert_almost_equal(xyz, expect_xyz)



"""Tigramite causal inference for time series."""

# Authors: Martin Rabel, Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

import numpy as np
from sklearn.neighbors import KNeighborsRegressor, KernelDensity
from functools import partial
import tigramite.toymodels.non_additive as toy_setup
from tigramite.causal_effects import CausalEffects
from tigramite.toymodels.non_additive import _Fct_on_grid, _Fct_smoothed



"""-------------------------------------------------------------------------------------------
-----------------------------   Helpers for Mixed Data Fitting   -----------------------------
-------------------------------------------------------------------------------------------"""


class MixedData:
    """Namescope for mixed-data helpers.

    Mostly for internal use.
    """

    @staticmethod
    def IsPurelyCategorical(vars_and_data):
        """Check if all variables are categorical

        Parameters
        ----------
        vars_and_data : dictionary< toy_setup.VariableDescription, np.array(N) >
            Samples with meta-data.

        Returns
        -------
        purely categorical : bool
            If all variables in vars_and_data are categorical, return true, otherwise false.
        """
        for var in vars_and_data.keys():
            if not var.Id().is_categorical:
                return False
        return True

    @staticmethod
    def _FixZeroDim(continuous_data, categorical_data):
        """[INTERNAL] Ensure consistent shape (0,N) for batches containing zero categorical or continuous variables"""
        if len(continuous_data) == 0:
            assert len(categorical_data) != 0
            continuous_data = np.empty([0, np.shape(categorical_data)[1]], dtype="float64")
            categorical_data = np.array(categorical_data)
        elif len(categorical_data) == 0:
            categorical_data = np.empty([0, np.shape(continuous_data)[1]], dtype="int32")
            continuous_data = np.array(continuous_data)
        else:
            continuous_data = np.array(continuous_data)
            categorical_data = np.array(categorical_data)
        return continuous_data, categorical_data

    @staticmethod
    def Get_data_via_var_ids(x, var_ids):
        """For a subset of variables, get separate batches for continuous and categorical data, and category-shape

        Parameters
        ----------
        x : dictionary< toy_setup.VariableDescription, np.array(N) >
            Samples and meta-data.
        var_ids : *iterable* <toy_setup.VariableDescription>
            Subset of variables to extract data for.

        Returns
        -------
        continuous_data : np.array( # continuous variables in var_ids, # samples )
            Continuous entries of the data.
        categorical_data : np.array( # categorical variables in var_ids, # samples )
            Categorical entries of the data.
        category_shape : list <uint>
            List of the number of categories for each categorical variable in var_ids in the same ordering.
        """
        category_shape = []
        categorical_data = []
        continuous_data = []
        for desc in var_ids:
            if desc.is_categorical:
                category_shape.append(desc.categories)
                categorical_data.append(x[desc])
            else:
                continuous_data.append(x[desc])
        return MixedData._FixZeroDim(continuous_data, categorical_data) + (category_shape,)

    @classmethod
    def Split_data_into_categorical_and_continuous(cls, x):
        """Get separate batches for continuous and categorical data, and category-shape

        (see Get_data_via_var_ids)
        """
        return cls.Get_data_via_var_ids(x, x.keys())

    @staticmethod
    def CategoryCount(category_shape):
        """Get total number of categories in shape.

        Parameters
        ----------
        category_shape : list<uint>
            Number of categories per variable.

        Returns
        -------
        total : uint
            Total number of categories in the product.
        """
        result = 1
        for c in category_shape:
            result *= c
        return result

    @staticmethod
    def Get_Len(continuous, categorical):
        """Extract number of samples from data"""
        if np.shape(continuous)[0] != 0:
            return np.shape(continuous)[1]
        else:
            assert np.shape(categorical)[0] != 0
            return np.shape(categorical)[1]

    @staticmethod
    def SimplifyIfTrivialVector(x):
        """Make shapes consistent."""
        result = np.squeeze(x)
        if np.shape(result) == ():
            return result[()]  # turn a scalar from array(x) of shape () to an actual scalar
        else:
            return result

    @classmethod
    def Call_map(cls, f, x):
        """Execute call consistently."""
        if hasattr(f, '__iter__'):
            result = []
            for coord in f:
                result.append(cls.Call_map(coord, x))
            return np.array(result)
        if callable(f):
            return cls.SimplifyIfTrivialVector(f(x))
        else:
            return cls.SimplifyIfTrivialVector(f.predict(x))


class FitProvider_Continuous_Default:
    r"""Helper for fitting continuous maps.

    See "Technical Appendix B" of the Causal Mediation Tutorial.

    Parameters
    ----------
    fit_map : *callable* (np.array(N, dim x), np.array(N))
        A callable, that given data (x,y) fits a regressor f s.t. y = f(x)
    fit_map_1d : *None* or *callable* (np.array(N, 1), np.array(N))
        overwrite fit_map in the case of a 1-dimensional domain.
    """

    def __init__(self, fit_map, fit_map_1d=None):
        self.fit_map = fit_map
        self.fit_map_1d = fit_map_1d

    @classmethod
    def UseSklearn(cls, neighbors=10, weights='uniform'):
        """Use an sci-kit learn KNeighborsRegressor

        Parameters
        ----------
        neighbors : uint
            The number of neighbors to consider
        weights : string
            Either 'uniform' or 'radius', see sci-kit learn.

        Returns
        -------
        K-Neighbors Fit-Provider : FitProvider_Continuous_Default
            Fit-Provider for use with FitSetup.
        """
        return cls(lambda x, y: KNeighborsRegressor(n_neighbors=neighbors, weights=weights).fit(x, y))

    @classmethod
    def UseSklearn_GC(cls, **kwargs):
        """Use an sci-kit learn GaussianProcessRegressor

        Parameters
        ----------
        ... : any-type
            forwarded to sci-kit

        Returns
        -------
        GC Fit-Provider : FitProvider_Continuous_Default
            Fit-Provider for use with FitSetup.
        """
        from sklearn.gaussian_process import GaussianProcessRegressor
        return cls(lambda x, y: GaussianProcessRegressor(**kwargs).fit(x, y))

    def Get_Fit_Continuous(self, x, y):
        """Produce a fit

        Parameters
        ----------
        x : np.array(N, dim x)
            Predictor values
        y : np.array(N)
            Training values

        Returns
        -------
        Continuous Fit : *callable* (x)
            The fitted predictor.
        """
        dim_of_domain = np.shape(x)[1]
        if dim_of_domain == 0:
            return lambda x_dim_zero: np.mean(y)
        elif dim_of_domain == 1 and self.fit_map_1d is not None:
            return self.fit_map_1d(x, y)  # use splines?
        else:
            return self.fit_map(x, y)


class FitProvider_Density_Default:
    r"""Helper for fitting continuous densities.

   See "Technical Appendix B" of the Causal Mediation Tutorial.

   Parameters
   ----------
   fit_density : *callable* (np.array(N, dim x))
       A callable, that given data (x) fits a density estimator p s.t. x ~ p
   """

    def __init__(self, fit_density):
        self.fit_density = fit_density

    @classmethod
    def UseSklearn(cls, kernel='gaussian', bandwidth=0.2):
        """Use an sci-kit learn KernelDensity

        Parameters
        ----------
        kernel : string
            E.g. 'gaussian', see sci-kit learn documentation.
        bandwidth : float
            See sci-kit learn documentation.

        Returns
        -------
        Density Estimator : FitProvider_Density_Default
            Fit-Provider for use with FitSetup.
        """
        return cls(lambda x: KernelDensity(kernel=kernel, bandwidth=bandwidth).fit(x))

    def Get_Fit_Density(self, x_train):
        """Produce a fit

        Parameters
        ----------
        x_train : np.array(N, dim x)
            Training samples.

        Returns
        -------
        Density Estimate : *callable* (x)
            The fitted predictor.
        """
        dim_of_domain = np.shape(x_train)[1]
        if dim_of_domain == 0:
            return lambda x_dim_zero: np.ones(np.shape(x_dim_zero)[0])
        elif np.shape(x_train)[0] > 0:
            model = self.fit_density(x_train)
            return lambda x_predict: np.exp(model.score_samples(x_predict))
        else:
            return lambda x_predict: np.zeros(np.shape(x_predict)[0])


class MixedMap:
    """Helper to evaluate fitted maps on mixed data.

    Used internally, should not normally be required in user-code.
    """
    error_value = float('nan')

    def __init__(self, maps, var_ids, dtype):
        self.maps = maps
        self.var_ids = var_ids
        self.dtype = dtype

    def predict(self, x):
        """Given x, predict y."""
        return self(x)

    def __call__(self, x):
        """Given x, predict y."""
        continuous_data, categorical_data, category_shape = MixedData.Get_data_via_var_ids(x, self.var_ids)
        category_count = MixedData.CategoryCount(category_shape)
        if category_count == 1:  # avoid errors in ravel_multi_index and improve performance by treating separately
            return MixedData.Call_map(self.maps, continuous_data.T)
        else:
            categories = np.ravel_multi_index(categorical_data, category_shape)
            result = np.empty(MixedData.Get_Len(continuous_data, categorical_data), dtype=self.dtype)
            for cX in range(category_count):
                filter_x = (categories == cX)
                if np.count_nonzero(filter_x) > 0:
                    # Avoid errors if asking for prediction on data with cX not occurring
                    # (Queries must be validated elsewhere)
                    filtered_x = continuous_data[:, filter_x]
                    if self.maps[cX] is not None:
                        result[filter_x] = MixedData.Call_map(self.maps[cX], filtered_x.T)
                    else:
                        result[filter_x] = self.error_value
            return result


class MixedMarkovKernel:
    """Helper to evaluate fitted densities on mixed data.

    Used internally, should not normally be required in user-code.
    """
    def __init__(self, transfers, var_ids, dtype, output_category_count):
        self.transfers = transfers
        self.var_ids = var_ids
        self.dtype = dtype
        self.output_category_count = output_category_count

    def predict(self, x):
        """Given np.array x, predict probability of each sample."""
        return self(x)

    def __call__(self, x):
        """Given np.array x, predict probability of each sample."""
        continuous_data, categorical_data, category_shape = MixedData.Get_data_via_var_ids(x, self.var_ids)
        category_count = MixedData.CategoryCount(category_shape)
        categories = None
        if category_count == 1:  # avoid errors in ravel_multi_index and improve performance by treating separately
            categories = np.zeros(np.shape(continuous_data)[1])
        else:
            categories = np.ravel_multi_index(categorical_data, category_shape)

        values = np.zeros([self.output_category_count, MixedData.Get_Len(continuous_data, categorical_data)])
        for cX in range(category_count):
            filter_x = (categories == cX)
            if np.count_nonzero(filter_x) > 0:
                # Avoid errors if asking for prediction on data with cX not occurring
                # (Queries must be validated elsewhere)
                filtered_x = continuous_data[:, filter_x]
                for cY in range(self.output_category_count):
                    transfer = self.transfers[cX][cY]
                    values[cY][filter_x] = transfer(filtered_x.T)
        normalize = np.sum(values, axis=0)
        return (values / normalize).T


class FitSetup:
    """Helper to fit mixed data.

   See "Technical Appendix B" of the Causal Mediation Tutorial.
    """
    def __init__(self, fit_map=FitProvider_Continuous_Default.UseSklearn(),
                 fit_density=FitProvider_Density_Default.UseSklearn()):
        self.fit_map = fit_map
        self.fit_density = fit_density

    def Fit(self, domain, target, dont_wrap_1d_y_in_vector=True):
        """Fit a mixed data mapping.

        See "Technical Appendix B" of the Causal Mediation Tutorial.
        """
        dont_wrap = dont_wrap_1d_y_in_vector and len(target) == 1
        result = {}
        for y, train in target.items():
            _next = None
            if y.is_categorical:
                _next = self.Fit_CategoricalTarget_MarkovKernel(domain, train, y.categories)
            else:
                _next = self.Fit_ContinuousTarget(domain, train, y.dimension)
            if dont_wrap:
                return _next
            else:
                result[y] = _next
        return result

    def Fit_ContinuousTarget(self, x, y, dim_y):
        """Fit a mixed-domain mapping with continuous target."""
        if dim_y == 1:
            return self.Fit_ContinuousTarget_1D(x, y)
        else:
            result = []
            for i in range(dim_y):
                result.append(self.Fit_ContinuousTarget_1D(x, y[:, i]))
            return result

    def Fit_ContinuousTarget_1D(self, x, y):
        """Fit a mixed-domain mapping with 1-dimensional continuous target."""
        continuous_data, categorical_data, category_shape = MixedData.Split_data_into_categorical_and_continuous(x)

        category_count = MixedData.CategoryCount(category_shape)
        if category_count == 1:
            # Avoid errors in np.ravel_multi_index and improve performance by treating this case separately
            _map = self.fit_map.Get_Fit_Continuous(continuous_data.T, y)
            return MixedMap(_map, var_ids=x.keys(), dtype=y.dtype)
        else:
            # Fit once per category
            category_index = np.ravel_multi_index(categorical_data, category_shape)
            maps = {}
            for cX in range(category_count):
                filter_x = (category_index == cX)
                filtered_y = y[filter_x]
                filtered_x = continuous_data[:, filter_x]
                if np.count_nonzero(filter_x) > 1:
                    maps[cX] = self.fit_map.Get_Fit_Continuous(filtered_x.T, filtered_y)
                else:
                    maps[cX] = None
            return MixedMap(maps, var_ids=x.keys(), dtype=y.dtype)

    def Fit_CategoricalTarget_MarkovKernel(self, x, y, y_category_count):
        """Fit a mixed-domain probabilistic mapping (a Markov-kernel) with categorical target."""
        continuous_data, categorical_data, category_shape = MixedData.Split_data_into_categorical_and_continuous(x)

        category_count = MixedData.CategoryCount(category_shape)
        category_index = None
        if category_count == 1:
            # Avoid errors in np.ravel_multi_index and improve performance by treating this case separately
            category_index = np.zeros(np.shape(continuous_data)[1])
        else:
            category_index = np.ravel_multi_index(categorical_data, category_shape)

        # Compute the transfer-matrix
        transfer_matrix = self._Fit_Enriched_TransferMatrix(continuous_data,
                                                            category_index, category_count,
                                                            y, y_category_count)
        return MixedMarkovKernel(transfer_matrix, var_ids=x, dtype=y.dtype, output_category_count=y_category_count)

    @staticmethod
    def _Fit_TransferMatrix(x_categorical, x_category_count, y_categorical, y_category_count):
        transfers = np.zeros([x_category_count, y_category_count])
        for cX in range(x_category_count):
            filter_x = (x_categorical == cX)
            filtered_y = y_categorical[filter_x]
            normalization = np.count_nonzero(filter_x)
            if normalization != 0:
                for cY in range(y_category_count):
                    transfers[cX, cY] = np.count_nonzero(filtered_y == cY) / normalization
            else:
                transfers[cX, cY] = MixedMap.error_value
        return transfers

    def _Fit_Enriched_TransferMatrix(self, x_continuous,
                                     x_categorical, x_category_count,
                                     y_categorical, y_category_count):
        transfers = []
        for cX in range(x_category_count):
            filter_x = (x_categorical == cX)
            normalization = np.count_nonzero(filter_x)
            transfers_from_cX = []
            transfers.append(transfers_from_cX)
            if normalization != 0:
                for cY in range(y_category_count):
                    # Use Bayes' Theorem to avoid fitting densities conditional on continuous variables
                    filter_both = np.logical_and(x_categorical == cX, y_categorical == cY)
                    assert len(filter_both.shape) == 1
                    p_y_given_discrete_x = np.count_nonzero(filter_both) / normalization
                    p_continuous_x_given_discrete_x_and_y = self.fit_density.Get_Fit_Density(
                        x_continuous[:, filter_both].T)
                    # Avoid weird bugs in lambda-captures combined with for-loop scopes by using "functools->partial"
                    transfers_from_cX.append(partial(self._EvalFromBayes, p_y_given_discrete_x=p_y_given_discrete_x,
                                                     p_continuous_x_given_discrete_x_and_y=p_continuous_x_given_discrete_x_and_y))
                    # transfers_from_cX.append(lambda x_c: self._EvalFromBayes(x_c, p_y_given_discrete_x,
                    #                                                        p_continuous_x_given_discrete_x_and_y))
            else:
                for cY in range(y_category_count):
                    transfers_from_cX.append(lambda x_c: np.full(np.shape(x_c)[0], MixedMap.error_value))
        return transfers

    @staticmethod
    def _EvalFromBayes(x_c, p_y_given_discrete_x, p_continuous_x_given_discrete_x_and_y):
        # P( y | x_c, x_d ) = P( x_c | x_d, y ) P( Y | x_d ) / normalization independent of y
        return p_continuous_x_given_discrete_x_and_y(x_c) * p_y_given_discrete_x





"""-------------------------------------------------------------------------------------------
--------------------------------   Natural Effect Estimation   -------------------------------
-------------------------------------------------------------------------------------------"""

class NaturalEffects_StandardMediationSetup:
    r"""Setup for estimating natural effects in a "standard" mediation (triangle) graph.

    Methods for the non-parametric estimation of mediation effects.
    For (more) general graphs, see NaturalEffects_GraphMediation.

    Actual fit-models can be chosen independently, for details see
    technical appendix B in the mediation-tutorial.

    See references and tigramite tutorial for an in-depth introduction.

    References
    ----------

    J. Pearl. Direct and indirect effects. Proceedings of the Seventeenth Conference
    on Uncertainty in Artificial intelligence, 2001.

    J. Pearl. Interpretation and identification of causal mediation. Psychological
    methods, 19(4):459, 2014.

    I. Shpitser and T. J. VanderWeele. A complete graphical criterion for the adjust-
    ment formula in mediation analysis. The international journal of biostatistics,
    7(1), 2011.

    Parameters
    ----------
    fit_setup : fit model
        A fit model to use internally, e.g. mixed_fit.FitSetup(). See there or technical
        appendix B of the tutorial.
    source : toy_setup.VariableDescription
        The (description of, e.g. toy_setup.ContinuousVariable()) the effect-source.
    target : toy_setup.VariableDescription
        The (description of, e.g. toy_setup.ContinuousVariable()) the effect-target.
    mediator : toy_setup.VariableDescription
        The (description of, e.g. toy_setup.ContinuousVariable()) the effect-mediator.
    data : dictionary ( keys=toy_setup.VariableDescription, values=np.array(N) )
        The data as map variable-description -> samples; e.g. toy_setup.world.Observables(),
        or toy_setup.VariablesFromDataframe( tigramite dataframe )
    """

    def __init__(self, fit_setup, source, target, mediator, data):
        self.fit_setup = fit_setup
        self.X = source
        self.Y = target
        self.M = mediator
        self.x = data[source]
        self.y = data[target]
        self.m = data[mediator]
        X = source
        Y = target
        M = mediator
        x = data[source]
        y = data[target]
        m = data[mediator]

        self._E_Y_XM = None
        self._P_Y_XM = None
        self._P_M_X = None

    def E_Y_XM_fixed_x_obs_m(self, x):
        """[internal] Provide samples Y( fixed x, observed m ) from fit of E[Y|X,M]

        Parameters
        ----------
        x : single value of same type as single sample for X (float, int, or bool)
            The fixed value of X.

        Returns
        -------
        Y( fixed x, observed m ) : np.array(N)
            Samples of Y estimated from observations and fit.
        """
        if self._E_Y_XM is None:
            self._E_Y_XM = self.fit_setup.Fit({self.X: self.x, self.M: self.m}, {self.Y: self.y})
        return self._E_Y_XM({self.X: np.full_like(self.x, x), self.M: self.m})

    def E_Y_XM_fixed_x_all_m(self, x):
        """For categorical M, provide samples Y( fixed x, m_i ) for all categories m_i of M from fit of E[Y|X,M]

        Parameters
        ----------
        x : single value of same type as single sample for X (float, int, or bool)
            The fixed value of X.

        Returns
        -------
        Y( fixed x, all  m_i ) : np.array(# of categories of M)
            Samples of Y estimated from fit for all categories of the mediator M.
        """
        assert self.M.is_categorical
        if self._E_Y_XM is None:
            self._E_Y_XM = self.fit_setup.Fit({self.X: self.x, self.M: self.m}, {self.Y: self.y})
        return self._E_Y_XM({self.X: np.full(self.M.categories, x), self.M: np.arange(self.M.categories)})

    def P_Y_XM_fixed_x_obs_m(self, x):
        """Provide for all samples of M the likelihood P( Y | fixed x, observed m ) from fit of P[Y|X,M]

        Parameters
        ----------
        x : single value of same type as single sample for X (float, int, or bool)
            The fixed value of X.

        Returns
        -------
        P( Y | fixed x, observed m ) : np.array(N, # categories of Y)
            Likelihood of each category of Y given x and observations of m from fit.
        """
        if self._P_Y_XM is None:
            self._P_Y_XM = self.fit_setup.Fit({self.X: self.x, self.M: self.m}, {self.Y: self.y})
        return self._P_Y_XM({self.X: np.full_like(self.x, x), self.M: self.m})

    def P_M_X(self, x):
        """Provide the likelihood P( M | fixed x ) from fit of P[M|X]

        Parameters
        ----------
        x : single value of same type as single sample for X (float, int, or bool)
            The fixed value of X.

        Returns
        -------
        P( M | fixed x ) : np.array( # categories of M )
            Likelihood of each category of M given x and observations of m from fit.
        """
        if self._P_M_X is None:
            self._P_M_X = self.fit_setup.Fit({self.X: self.x}, {self.M: self.m})
        return self._P_M_X({self.X: x})

    def _ValidateRequest(self, change_from, change_to):
        """Sanity-check parameters passed to effect-estimation

        Parameters
        ----------
        cf. NDE, NIE

        Throws
        ------
        Raises and exception if parameters are not meaningful
        """
        if self.X.is_categorical:
            # both_int = isinstance(change_from, int) and isinstance(change_to, int) does not work with numpy
            both_int = change_from % 1 == 0.0 and change_to % 1 == 0.0
            if not both_int:
                raise Exception("Categorical variables can only be intervened to integer values.")
            if change_from >= self.X.categories or change_to >= self.X.categories or change_from < 0 or change_to < 0:
                raise Exception("Intervention on categorical variable was outside of valid category-range.")

    def NDE(self, change_from, change_to, fct_of_nde=None):
        """Compute Natural Direct Effect (NDE)

        Parameters
        ----------
        change_from : single value of same type as single sample for X (float, int, or bool)
            Reference-value to which X is set by intervention in the world seen by the mediator.

        change_to : single value of same type as single sample for X (float, int, or bool)
            Post-intervention-value to which X is set by intervention in the world seen by the effect (directly).

        fct_of_nde : *callable* or None
            Also in the case of a continuous Y the distribution, not just the expectation is identified.
            However, a density-fit is not usually reasonable to do in practise. However, instead of
            computing E[Y] can compute E[f(Y)] efficiently for any f. Assume f=id if None.

        Throws
        ------
        Raises and exception if parameters are not meaningful

        Returns
        -------
        NDE : If Y is categorical -> np.array( # categories Y, 2 )
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.

        NDE : If Y is continuous -> float
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.
        """
        self._ValidateRequest(change_from, change_to)
        if self.Y.is_categorical:
            assert fct_of_nde is None, "Categorical estimate returns full density-estimate anyway."
            return self._NDE_categorical_target(change_from, change_to)
        else:
            return self._NDE_continuous_target(change_from, change_to, fct_of_nde)

    def NIE(self, change_from, change_to):
        """Compute Natural Indirect Effect (NIE)

        Parameters
        ----------
        change_from : single value of same type as single sample for X (float, int, or bool)
           Reference-value to which X is set by intervention in the world seen by the effect (directly).

        change_to : single value of same type as single sample for X (float, int, or bool)
            Post-intervention-value to which X is set by intervention in the world seen by the mediator.

        Throws
        ------
        Raises and exception if parameters are not meaningful

        Returns
        -------
        NIE : If Y is categorical -> np.array( # categories Y, 2 )
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by M from change_from to change_to, while keeping the value as (directly) seen by Y,
            as if X remained at change_from.

        NIE : If Y is continuous -> float
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by M from change_from to change_to, while keeping the value as (directly) seen by Y,
            as if X remained at change_from.
        """
        self._ValidateRequest(change_from, change_to)
        if self.Y.is_categorical:
            return self._NIE_categorical_target(change_from, change_to)
        else:
            return self._NIE_continuous_target(change_from, change_to)

    def _NDE_continuous_target(self, change_from, change_to, fct_of_nde=None):
        """Compute NDE (continuous Y)

        See 'NDE' above.

        Computed from mediation-formula
        (see [Pearl 2001] or [Shpitser, VanderWeele], see references above)
        by "double"-regression.
        """
        difference = self.E_Y_XM_fixed_x_obs_m(change_to) - self.E_Y_XM_fixed_x_obs_m(change_from)

        if fct_of_nde is not None:
            difference = fct_of_nde(difference)

        # sklearn predicts nan if too far from data ... (but might be irrelevant, check separately)
        valid_samples = np.isfinite(difference)
        Difference = toy_setup.ContinuousVariable()
        X = self.X
        x = self.x
        E_Difference_X = self.fit_setup.Fit({X: x[valid_samples]}, {Difference: difference[valid_samples]})
        return E_Difference_X({X: [change_from]})

    def _NDE_categorical_target_full_density(self, actual_value, counterfactual_value):
        """Compute NDE as full density (categorical Y)

        See 'NDE' above.

        Computed from mediation-formula
        (see [Pearl 2001] or [Shpitser, VanderWeele], see references above)
        by "double"-regression.

        Note: According to (see p.13)
        [Shpitser, VanderWeele: A Complete Graphical Criterion for theAdjustment Formula in Mediation Analysis]
        not just the expectation-value, but the full counterfactual distribution can be obtained via mediation-formula.
        """
        P_Y = toy_setup.ContinuousVariable(dimension=self.Y.categories)
        P_Y_samples = self.P_Y_XM_fixed_x_obs_m(actual_value)
        P_Y_X = self.fit_setup.Fit({self.X: self.x}, {P_Y: P_Y_samples})

        return MixedData.Call_map(P_Y_X, {self.X: [counterfactual_value]})

    def _NDE_categorical_target(self, actual_value, counterfactual_value):
        """Compute NDE by evaluating density (categorical Y)

        See 'NDE' and '_NDE_categorical_target_full_density' above.
        """
        # returns (counterfactual probabilities, total effect)
        return np.array([self._NDE_categorical_target_full_density(counterfactual_value, actual_value),
                         self._NDE_categorical_target_full_density(actual_value, actual_value)]).T

    def _NIE_continuous_target(self, change_from, change_to):
        """Compute NIE (continuous Y)

        See 'NIE' above.

        Computed from mediation-formula
        (see [Pearl 2001] or [Shpitser, VanderWeele], see references above)
        by "double"-regression.

        If M is categorical, after fixing X=x, the fitted P( Y | X=x, M=m ), is actually
        categorical (a transfer matrix), because it takes values only in
        im( P ) = { P( Y | X=x, M=m_0 ), ...,  P( Y | X=x, M=m_k ) } where
        m_0, ..., m_k are the categories of M. This is clearly a finite set.
        Since the distribution over this finite subset of the continuous Val(Y)
        is very non-gaussian, "max likelihood by least square estimation" can fail horribly.
        Hence we fit a transfer-matrix instead.
        """
        X = self.X
        M = self.M
        x = self.x
        m = self.m

        if M.is_categorical:
            # if the image of the mapping in double-regression is finite, use a density fit instead
            p_M_X01 = self.P_M_X([change_from, change_to])
            E_YX0 = self.E_Y_XM_fixed_x_all_m(change_from)
            # sum over finite M:
            return np.dot(E_YX0, p_M_X01[1] - p_M_X01[0])

        else:
            Estimate = toy_setup.ContinuousVariable()
            y_at_original_x = self.E_Y_XM_fixed_x_obs_m(change_from)
            ModifiedY = self.fit_setup.Fit({X: x}, {Estimate: y_at_original_x})
            return (MixedData.Call_map(ModifiedY, {X: [change_to]})
                    - MixedData.Call_map(ModifiedY, {X: [change_from]}))

    def _NIE_categorical_target(self, change_from, change_to):
        """Compute NIE (continuous Y)

        See 'NIE' above.

        Computed from mediation-formula
        (see [Pearl 2001] or [Shpitser, VanderWeele], see references above)
        by "double"-regression.

        Similar to before (see _NIE_continuous_target), treat categorical M differently.
        """
        X = self.X
        M = self.M
        x = self.x
        m = self.m
        py_at_original_x = self.P_Y_XM_fixed_x_obs_m(change_from)

        if M.is_categorical:
            # if the image of the mapping in double-regression is finite, use a density fit instead
            p_M_X01 = self.P_M_X([change_from, change_to])
            # sum over finite M:
            result = np.zeros([self.Y.categories, 2])  # 2 is for TE & CF
            for cM in range(M.categories):
                result += np.outer(np.mean(py_at_original_x[m == cM], axis=0), p_M_X01[:, cM])
            return result

        else:
            Estimate = toy_setup.ContinuousVariable(dimension=self.Y.categories)
            ModifiedY = self.fit_setup.Fit({X: x}, {Estimate: py_at_original_x})
            return np.array([MixedData.Call_map(ModifiedY, {X: [change_to]}),
                             MixedData.Call_map(ModifiedY, {X: [change_from]})]).T

    def NDE_grid(self, list_of_points, cf_delta=0.5, normalize_by_delta=False, fct_of_nde=None):
        """Compute NDE as grided (unsmoothed) function

        Parameters
        ----------
        list_of_points : np.array( K )
            List of reference-values at which to estimate NDEs

        cf_delta : float
            The change from reference-value to effect-value (change_from=reference, change_to=ref + delta)

        normalize_by_delta : bool
            Normalize the effect by dividing by cf_delta.

        fct_of_nde : *callable* or None
            Also in the case of a continuous Y the distribution, not just the expectation is identified.
            However, a density-fit is not usually reasonable to do in practise. However, instead of
            computing E[Y] can compute E[f(Y)] efficiently for any f. Assume f=id if None.

        Throws
        ------
        Raises and exception if parameters are not meaningful

        Returns
        -------
        NDE : If Y is categorical -> np.array( K = # grid points, # categories Y, 2 )
            For each grid-point:
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.

        NDE : If Y is continuous -> np.array( K = # grid points )
            For each grid-point:
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.
        """
        return _Fct_on_grid(self.NDE, list_of_points, cf_delta, normalize_by_delta, fct_of_nde=fct_of_nde)

    def NIE_grid(self, list_of_points, cf_delta=0.5, normalize_by_delta=False):
        """Compute NIE as grided (unsmoothed) function

        Parameters
        ----------
        list_of_points : np.array( K )
            List of reference-values at which to estimate NIEs

        cf_delta : float
            The change from reference-value to effect-value (change_from=reference, change_to=ref + delta)

        normalize_by_delta : bool
            Normalize the effect by dividing by cf_delta.

        Throws
        ------
        Raises and exception if parameters are not meaningful

        Returns
        -------
        NIE : If Y is categorical -> np.array( K = # grid points, # categories Y, 2 )
            For each grid-point:
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by M from change_from to change_to, while keeping the value of X "seen" (directly)
            by Y as if X remained at change_from.

        NIE : If Y is continuous -> np.array( K = # grid points )
            For each grid-point:
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by M from change_from to change_to, while keeping the value of X "seen" (directly)
            by Y as if X remained at change_from.
        """
        return _Fct_on_grid(self.NIE, list_of_points, cf_delta, normalize_by_delta)

    def NDE_smoothed(self, min_x, max_x, cf_delta=0.5, steps=100, smoothing_gaussian_sigma_in_steps=5,
                     normalize_by_delta=False, fct_of_nde=None):
        """Compute NDE as smoothed function

        Parameters
        ----------
        min_x : float
            Lower bound of interval on which reference-values for X are taken

        max_x : float
            Upper bound of interval on which reference-values for X are taken

        cf_delta : float
            The change from reference-value to effect-value (change_from=reference, change_to=ref + delta)

        steps : uint
            Number of intermediate values to compute in the interval [min_x, max_x]

        smoothing_gaussian_sigma_in_steps : uint
            The width of the Gauß-kernel used for smoothing, given in steps.

        normalize_by_delta : bool
            Normalize the effect by dividing by cf_delta.

        fct_of_nde : *callable* or None
            Also in the case of a continuous Y the distribution, not just the expectation is identified.
            However, a density-fit is not usually reasonable to do in practise. However, instead of
            computing E[Y] can compute E[f(Y)] efficiently for any f. Assume f=id if None.

        Throws
        ------
        Raises and exception if parameters are not meaningful

        Returns
        -------
        NDE : If Y is categorical -> np.array( # steps, # categories Y, 2 )
            For each grid-point:
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.

        NDE : If Y is continuous -> np.array( # steps )
            For each grid-point:
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping M as if X remained at change_from.
        """
        return _Fct_smoothed(self.NDE, min_x, max_x, cf_delta, steps, smoothing_gaussian_sigma_in_steps,
                             normalize_by_delta, fct_of_nde=fct_of_nde)

    def NIE_smoothed(self, min_x, max_x, cf_delta=0.5, steps=100, smoothing_gaussian_sigma_in_steps=5,
                     normalize_by_delta=False):
        """Compute NIE as smoothed function

        Parameters
        ----------
        min_x : float
            Lower bound of interval on which reference-values for X are taken

        max_x : float
            Upper bound of interval on which reference-values for X are taken

        cf_delta : float
            The change from reference-value to effect-value (change_from=reference, change_to=ref + delta)

        steps : uint
            Number of intermediate values to compute in the interval [min_x, max_x]

        smoothing_gaussian_sigma_in_steps : uint
            The width of the Gauß-kernel used for smoothing, given in steps.

        normalize_by_delta : bool
            Normalize the effect by dividing by cf_delta.

        Throws
        ------
        Raises and exception if parameters are not meaningful

        Returns
        -------
        NIE : If Y is categorical -> np.array( # steps, # categories Y, 2 )
            For each grid-point:
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by M from change_from to change_to, while keeping the value of X "seen" (directly)
            by Y as if X remained at change_from.

        NDE : If Y is continuous -> np.array( # steps )
            For each grid-point:
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by M from change_from to change_to, while keeping the value of X "seen" (directly)
            by Y as if X remained at change_from.
        """
        return _Fct_smoothed(self.NIE, min_x, max_x, cf_delta, steps, smoothing_gaussian_sigma_in_steps,
                             normalize_by_delta)


class NaturalEffects_GraphMediation:
    r"""Setup for estimating natural effects in a (general) causal graph.

    Methods for the non-parametric estimation of mediation effects by adjustment.

    Actual fit-models can be chosen independently, for details see
    technical appendix B in the mediation-tutorial.

    See references and tigramite tutorial for an in-depth introduction.

    References
    ----------

    J. Pearl. Direct and indirect effects. Proceedings of the Seventeenth Conference
    on Uncertainty in Artificial intelligence, 2001.

    J. Pearl. Interpretation and identification of causal mediation. Psychological
    methods, 19(4):459, 2014.

    I. Shpitser and T. J. VanderWeele. A complete graphical criterion for the adjust-
    ment formula in mediation analysis. The international journal of biostatistics,
    7(1), 2011.

    Parameters
    ----------
    graph : np.array( [N, N] or [N, N, tau_max+1] depending on graph_type ) of 3-character patterns
        The causal graph, see 'Causal Effects' tutorial. E.g. returned by causal discovery method
        (see "Tutorials/Causal Discovery/CD Overview") or by a toymodel (see toy_setup.Model.GetGroundtruthGraph or
        the Mediation tutorial).
    graph_type : string
        The type of graph, tested for 'dag' and 'stationary_dag' (time-series). See 'Causal Effects' tutorial.
    tau_max : uint
        Maximum lag to be considered (can be relevant for adjustment sets, passed to 'Causal Effects' class).
    fit_setup : fit model
        A fit model to use internally, e.g. mixed_fit.FitSetup(). See there or technical
        appendix B of the tutorial.
    observations_data : dictionary ( keys=toy_setup.VariableDescription, values=np.array(N) )
        The data as map variable-description -> samples; e.g. toy_setup.world.Observables(),
        or toy_setup.VariablesFromDataframe( tigramite dataframe )
    effect_source : toy_setup.VariableDescription or (idx, -lag)
        The (description of, e.g. toy_setup.ContinuousVariable()) the effect-source.
    effect_target : toy_setup.VariableDescriptionor (idx, -lag)
        The (description of, e.g. toy_setup.ContinuousVariable()) the effect-target.
    blocked_mediators : iterable of Variable-descriptions or 'all'
        Which mediators to 'block' (consider indirect), *un*\ blocked mediators are considered as
        contributions to the *direct* effect.
    adjustment_set : iterable of Variable-descriptions or 'auto'
        Adjustment-set to use. Will be validated if specified explicitly, if 'auto', will try
        to use an 'optimal' set, fall back to [Perkovic et al]'s adjustment-set (which should always
        work if single-set adjustment as in [Shpitser, VanderWeele] is possible; this follows
        from combining results of [Shpitser, VanderWeele] and [Perkovic et al]).
        See 'Causal Effects' and its tutorial for more info and references on (optimal) adjustment.
    only_check_validity : bool
        If True, do not set up an estimator, only check if an optimal adjustment-set exists (or the
        explicitly specified one is valid). Call this.Valid() to extract the result.
    fall_back_to_total_effect : bool
        If True, if no mediators are blocked, use mediation implementation to estimate the total effect.
        In this case, estimating the total effect through the 'Causal Effects' class might be easier,
        however, for comparison to other estimates, using this option might yield more consistent results.
    _internal_provide_cfx : *None* or tigramite.CausalEffects
        Set to None. Used when called from CausalMediation, which already has a causal-effects class.
    enable_dataframe_based_preprocessing : bool
        Enable (and enforce) data-preprocessing through the tigramite::dataframe, makes missing-data
        and other features available to the mediation analysis. Custom (just in time) handling
        of missing data might be more sample-efficient.
    """

    def __init__(self, graph, graph_type, tau_max, fit_setup, observations_data, effect_source, effect_target,
                 blocked_mediators="all", adjustment_set="auto", only_check_validity=False,
                 fall_back_to_total_effect=False, _internal_provide_cfx=None, enable_dataframe_based_preprocessing=True):

        data = toy_setup.DataHandler(observations_data, dataframe_based_preprocessing=enable_dataframe_based_preprocessing)

        self.Source = data.GetVariableAuto(effect_source, "Source")
        self.Target = data.GetVariableAuto(effect_target, "Target")

        if blocked_mediators != "all":
            blocked_mediators = data.GetVariablesAuto(blocked_mediators, "Mediator")
        if adjustment_set != "auto":
            adjustment_set = data.GetVariablesAuto(adjustment_set, "Adjustment")

        X = data[self.Source]
        Y = data[self.Target]

        # Use tigramite's total effect estimation utility to help generate adjustment-sets and mediators
        cfx_xy = _internal_provide_cfx if _internal_provide_cfx is not None else\
            CausalEffects(graph, graph_type=graph_type, X=[X], Y=[Y])

        # ----- MEDIATORS -----
        # Validate "blocked mediators" are actually mediators, or find "all"
        all_mediators = blocked_mediators == "all"
        if all_mediators:
            M = cfx_xy.M
            blocked_mediators = data.ReverseLookupMulti(M, "Mediator")
        else:
            M = data[blocked_mediators]
            if not set(M) <= set(cfx_xy.M):
                raise Exception("Blocked mediators, if specified, must actually be mediators, try using"
                                "set-intersection with causal_effects_instance_xy.M instead.")
        if len(M) == 0 and not fall_back_to_total_effect:
            raise Exception("There are no mediators, use total-effect estimation instead or set "
                            "fall_back_to_total_effect=True!")

        # ----- ADJUSTMENT -----
        # Use tigramite's total effect estimation utility to help validate adjustment-sets and mediators
        cfx_xm = CausalEffects(graph, graph_type=graph_type, X=[X], Y=M)
        cfx_xm_y = CausalEffects(graph, graph_type=graph_type, X=[X] + list(M), Y=[Y])

        def valid(S):
            return cfx_xm._check_validity(S) and cfx_xm_y._check_validity(S)

        adjustment_set_auto = (adjustment_set == "auto")

        if adjustment_set_auto:
            # first try optimal set (for small cf_delta, optimality for the causal effect should imply
            # optimality for the nde by continuity for the estimator variance)
            Z = cfx_xy.get_optimal_set()

            if not valid(Z) and adjustment_set_auto:
                # fall back to adjust, which should work if any single adjustmentset works
                Z = cfx_xy._get_adjust_set()

            adjustment_set = data.ReverseLookupMulti(Z, "Adjustment")

        else:
            Z = data[adjustment_set]

        self.valid = valid(Z)
        if only_check_validity:
            return

        # Output appropriate error msgs if not valid
        if not self.valid:
            if adjustment_set_auto:
                raise Exception(
                    "The graph-effect you are trying to estimate is not identifiable via one-step adjustment, "
                    "try using a different query "
                    "or refine the causal graph by expert knowledge. "
                    "For the implemented method, there must be an adjustment-set, valid for both X u M -> Y and "
                    "X -> Y. If such a set exists, Perkovic's Adjust(X,Y) is valid, which was tried as "
                    "fallback because adjustment-set='auto' was used.")
            else:
                raise Exception(
                    "The adjustment-set specified is not valid for one-step adjustment, "
                    "try using or a different adjustment-set (or set it to 'auto'), a different query "
                    "or refine the causal graph by expert knowledge. "
                    "For the implemented method, there must be an adjustment-set, valid for both X u M -> Y and "
                    "X -> Y. If such a set exists, Perkovic's Adjust(X,Y) is valid, which is tried as "
                    "fallback if adjustment-set='auto' is used.")

        
        # lock in mediators and adjustment for preprocessing
        self.BlockedMediators = data.ReverseLookupMulti(M)
        self.AdjustmentSet = data.ReverseLookupMulti(Z)

        # ----- STORE RESULTS ON INSTANCE -----
        self.X, self.sources = data.Get("Source", [X], tau_max=tau_max)
        self.X = self.X[0]  # currently univariate anyway
        self.Y, self.targets = data.Get("Target", [Y], tau_max=tau_max)
        self.Y = self.Y[0]
        if len(M) > 0:
            self.M_ids, self.mediators = data.Get("Mediator", M, tau_max=tau_max)
        else:
            self.M_ids = None
            self.mediators = {}


        if len(Z) > 0:
            self.Z_ids, self.adjustment = data.Get("Adjustment", Z, tau_max=tau_max)
        else:
            self.Z_ids = None
            self.adjustment = {}


        self.fit_setup = fit_setup
        self._E_Y_XMZ = None
        self._P_Y_XMZ = None

    def E_Y_XMZ_fixed_x_obs_mz(self, x):
        """Provide samples Y( fixed x, observed m, z ) from fit of E[Y|X,M,Z]

        Parameters
        ----------
        x : single value of same type as single sample for X (float, int, or bool)
            The fixed value of X.

        Returns
        -------
        Y( fixed x, observed m, z ) : np.array(N)
            Samples of Y estimated from observations and fit.
        """
        assert not self.Y.is_categorical
        if self._E_Y_XMZ is None:
            self._E_Y_XMZ = self.fit_setup.Fit({**self.sources, **self.mediators, **self.adjustment}, self.targets)

        return self._E_Y_XMZ({self.X: np.full_like(self.sources[self.X], x),
                              **self.mediators, **self.adjustment})

    def P_Y_XMZ_fixed_x_obs_mz(self, x):
        """Provide for all samples of M, Z the likelihood P( Y | fixed x, observed m, z ) from fit of P[Y|X,M,Z]

        Parameters
        ----------
        x : single value of same type as single sample for X (float, int, or bool)
            The fixed value of X.

        Returns
        -------
        P( Y | fixed x, observed m, z ) : np.array(N, # categories of Y)
            Likelihood of each category of Y given x and observations of m, z from fit.
        """
        assert self.Y.is_categorical
        if self._P_Y_XMZ is None:
            self._P_Y_XMZ = self.fit_setup.Fit({**self.sources, **self.mediators, **self.adjustment}, self.targets)
        return self._P_Y_XMZ({self.X: np.full_like(self.sources[self.X], x),
                              **self.mediators, **self.adjustment})

    def Valid(self):
        """Get validity of adjustment-set

        Returns
        -------
        Valid : bool
            Validity of adjustment set. (see constructor-parameter 'only_check_validity')
        """
        return self.valid

    def NDE(self, change_from, change_to):
        """Compute Natural Direct Effect (NDE)

        Parameters
        ----------
        change_from : single value of same type as single sample for X (float, int, or bool)
            Reference-value to which X is set by intervention in the world seen by the (blocked) mediators.

        change_to : single value of same type as single sample for X (float, int, or bool)
            Post-intervention-value to which X is set by intervention in the world seen by the effect (directly).

        Throws
        ------
        Raises and exception if parameters are not meaningful or if adjustment-set is not valid.

        Returns
        -------
        NDE : If Y is categorical -> np.array( # categories Y, 2 )
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.

        NDE : If Y is continuous -> float
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.
        """
        if not self.valid:
            raise Exception("Valid adjustment-set is required!")
        if not (self.X.ValidValue(change_from) and self.X.ValidValue(change_to)):
            raise Exception("NDE change must be at valid values of the source-variable (e.g. categorical, and within "
                            "the range [0,num-categories).")
        if self.Y.is_categorical:
            return np.array([self._NDE_categorical_target_full_density(change_to, change_from),
                             self._NDE_categorical_target_full_density(change_from, change_from)]).T
        else:
            return self._NDE_continuous_target(change_from, change_to)

    def _NDE_continuous_target(self, change_from, change_to):
        """Compute NDE (continuous Y)

        See 'NDE' above.

        Computed from mediation-formula
        (see [Pearl 2001] or [Shpitser, VanderWeele], see references above)
        by "triple"-regression.
        """
        difference = (self.E_Y_XMZ_fixed_x_obs_mz(change_to)
                      - self.E_Y_XMZ_fixed_x_obs_mz(change_from))

        Difference = toy_setup.ContinuousVariable()
        E_Difference_X = self.fit_setup.Fit({**self.sources, **self.adjustment},
                                            {Difference: difference})

        E_NDE_per_c = E_Difference_X({self.X: np.full_like(self.sources[self.X], change_from), **self.adjustment})

        return np.mean(E_NDE_per_c)

    def _NDE_categorical_target_full_density(self, cf_x, reference_x):
        """Compute NDE as full density (categorical Y)

        See 'NDE' above.

        Computed from mediation-formula
        (see [Pearl 2001] or [Shpitser, VanderWeele], see references above)
        by "double"-regression.

        Note: According to (see p.13)
        [Shpitser, VanderWeele: A Complete Graphical Criterion for theAdjustment Formula in Mediation Analysis]
        not just the expectation-value, but the full counterfactual distribution can be obtained via mediation-formula.

        Note: If all M and Z are categorical, after fixing X=x, the fitted P( Y | X=x, M=m ), is actually
        categorical (a transfer matrix), because it takes values only in
        im( P ) = { P( Y | X=x, M=m_0 ), ...,  P( Y | X=x, M=m_k ) } where
        m_0, ..., m_k are the categories of M. This is clearly a finite set.
        Since the distribution over this finite subset of the continuous Val(Y)
        is very non-gaussian, "max likelihood by least square estimation" can fail horribly.
        Hence we fit a transfer-matrix instead.
        """
        p_y_values = self.P_Y_XMZ_fixed_x_obs_mz(cf_x)

        if (MixedData.IsPurelyCategorical({**self.mediators, **self.adjustment})
                and len({**self.mediators, **self.adjustment}) > 0):
            # If there are mediators or adjustment
            # and they are purely categorical, then the mapping (M u Z) -> P_Y
            # has finite image, treating it as categorical gives better results

            # different numpy-versions behave differently wrt this call:
            # https://numpy.org/devdocs/release/2.0.0-notes.html#np-unique-return-inverse-shape-for-multi-dimensional-inputs
            # see also https://github.com/numpy/numpy/issues/26738
            labels_y, transformed_y_numpy_version_dependent = np.unique(p_y_values, return_inverse=True, axis=0)
            transformed_y = transformed_y_numpy_version_dependent.squeeze()

            P_Y = toy_setup.CategoricalVariable(categories=labels_y)
            P_P_Y_xz = self.fit_setup.Fit({**self.sources, **self.adjustment}, {P_Y: transformed_y})

            C_NDE_per_c = MixedData.Call_map(P_P_Y_xz,
                                                       {self.X: np.full_like(self.sources[self.X], reference_x),
                                                        **self.adjustment})
            print(labels_y)
            print(C_NDE_per_c.shape)
            print(labels_y.shape)
            P_NDE_per_c = np.matmul(C_NDE_per_c, labels_y)
            return np.mean(P_NDE_per_c, axis=0)  # Axis 1 is p of different categories

        else:
            P_Y = toy_setup.ContinuousVariable(dimension=self.Y.categories)
            E_P_Y_xz = self.fit_setup.Fit({**self.sources, **self.adjustment}, {P_Y: p_y_values})

            P_NDE_per_c = MixedData.Call_map(E_P_Y_xz,
                                                       {self.X: np.full_like(self.sources[self.X], reference_x),
                                                        **self.adjustment})

            return np.mean(P_NDE_per_c, axis=1)  # Axis 0 is p of different categories

    def NDE_smoothed(self, min_x, max_x, cf_delta=0.5, steps=100, smoothing_gaussian_sigma_in_steps=5,
                     normalize_by_delta=False):
        """Compute NDE as smoothed function

        Parameters
        ----------
        min_x : float
            Lower bound of interval on which reference-values for X are taken

        max_x : float
            Upper bound of interval on which reference-values for X are taken

        cf_delta : float
            The change from reference-value to effect-value (change_from=reference, change_to=ref + delta)

        steps : uint
            Number of intermediate values to compute in the interval [min_x, max_x]

        smoothing_gaussian_sigma_in_steps : uint
            The width of the Gauß-kernel used for smoothing, given in steps.

        normalize_by_delta : bool
            Normalize the effect by dividing by cf_delta.

        Throws
        ------
        Raises and exception if parameters are not meaningful, adjustment-set is not valid or
        normalization requested makes no sense (normalizing probabilites by delta).

        Returns
        -------
        NDE : If Y is categorical -> np.array( # steps, # categories Y, 2 )
            For each grid-point:
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.

        NDE : If Y is continuous -> np.array( # steps )
            For each grid-point:
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.
        """
        if self.Y.is_categorical and normalize_by_delta:
            raise Exception("Do not normalize categorical output-probabilities by delta. (They are probabilities, "
                            "so normalizing them in this way makes no sense.) Normalize the difference instead.")
        return _Fct_smoothed(self.NDE, min_x, max_x, cf_delta, steps, smoothing_gaussian_sigma_in_steps,
                             normalize_by_delta)

    def PrintInfo(self, detail=1):
        """Print info about estimator.

        Helper to quickly print blocked mediators, adjustment-set used and source->target.
        """
        print(f"Estimator for the effect of {self.Source.Info(detail)} on {self.Target.Info(detail)}")
        self.PrintMediators(detail)
        self.PrintAdjustmentSet(detail)

    def PrintMediators(self, detail=1):
        """ Print info about blocked mediators. """
        print("Blocked Mediators:")
        for m in self.BlockedMediators:
            print(" - " + m.Info(detail))

    def PrintAdjustmentSet(self, detail=1):
        """ Print info about adjustment-set. """
        print("Adjustment Set:")
        for z in self.AdjustmentSet:
            print(" - " + z.Info(detail))




"""-------------------------------------------------------------------------------------------
-------------------------------   Expose Tigramite Interface   -------------------------------
-------------------------------------------------------------------------------------------"""



class CausalMediation(CausalEffects):
    """Non-linear, non-additive causal mediation analysis.

    See the tutorial on Causal Mediation.

    Extends the tigramite.CausalEffects class by natural-effect estimation for counter-factual mediation analysis.
    Effects are estimated by adjustment, where adjustment-sets can be generated automatically (if they exist).

    Actual fit-models can be chosen independently, for details see
    technical appendix B in the mediation-tutorial.

    See references and tigramite tutorial for an in-depth introduction.

    References
    ----------
    J. Pearl. Direct and indirect effects. Proceedings of the Seventeenth Conference
    on Uncertainty in Artificial intelligence, 2001.

    J. Pearl. Interpretation and identification of causal mediation. Psychological
    methods, 19(4):459, 2014.

    I. Shpitser and T. J. VanderWeele. A complete graphical criterion for the adjust-
    ment formula in mediation analysis. The international journal of biostatistics,
    7(1), 2011.

    Parameters
    ----------
    graph : np.array( [N, N] or [N, N, tau_max+1] depending on graph_type ) of 3-character patterns
        The causal graph, see 'Causal Effects' tutorial. E.g. returned by causal discovery method
        (see "Tutorials/Causal Discovery/CD Overview") or by a toymodel (see toy_setup.Model.GetGroundtruthGraph or
        the Mediation tutorial).
    graph_type : string
        The type of graph, tested for 'dag' and 'stationary_dag' (time-series). See 'Causal Effects' tutorial.
    X : (idx, -lag)
        Index of the effect-source.
    Y : (idx, -lag)
        Index of the effect-target.
    S : None
        Reserved. Must be None in current version.
    hidden_variables : None
        Reserved. Must be None in current version.
    verbosity : uint
        Tigramite.CausalEffects verbosity setting.
    """
    def __init__(self, graph, graph_type, X, Y, S=None, hidden_variables=None, verbosity=0):
        super().__init__(graph, graph_type, X, Y, S, hidden_variables, False, verbosity)
        self.BlockedMediators = None
        self.MediationEstimator = None
        assert hidden_variables is None

    def fit_natural_direct_effect(self, dataframe, mixed_data_estimator=FitSetup(),
                                  blocked_mediators='all', adjustment_set='auto',
                                  use_mediation_impl_for_total_effect_fallback=False,
                                  enable_dataframe_based_preprocessing=True):
        """Fit a natural direct effect.

        Parameters
        ----------
        dataframe : tigramite.Dataframe
            Observed data.
        mixed_data_estimator : mixed_fit.FitSetup
            The fit-configuration to use. See mixed_fit.FitSetup and the Mediation tutorial, Appendix B.
        blocked_mediators : 'all' or *iterable* of < (idx, -lag) >
            Which mediators to 'block' (consider indirect), *un*\ blocked mediators are considered as
            contributions to the *direct* effect.
        adjustment_set : 'auto' or None or *iterable* < (idx, -lag) >
            Adjustment-set to use. Will be validated if specified explicitly, if 'auto' or None, will try
            to use an 'optimal' set, fall back to [Perkovic et al]'s adjustment-set (which should always
            work if single-set adjustment as in [Shpitser, VanderWeele] os possible; this follows
            from combining results of [Shpitser, VanderWeele] and [Perkovic et al]).
            See 'Causal Effects' and its tutorial for more info and references on (optimal) adjustment.
        use_mediation_impl_for_total_effect_fallback : bool
            If True, if no mediators are blocked, use mediation implementation to estimate the total effect.
            In this case, estimating the total effect through the 'Causal Effects' class might be easier,
            however, for comparison to other estimates, using this option might yield more consistent results.        
        enable_dataframe_based_preprocessing : bool
            Enable (and enforce) data-preprocessing through the tigramite::dataframe, makes missing-data
            and other features available to the mediation analysis. Custom (just in time) handling
            of missing data might be more sample-efficient.

        Returns
        -------
        estimator : NaturalEffects_GraphMediation
            Typically, use predict_natural_direct_effect or predict_natural_direct_effect_function
            to use the fitted estimator.
            The internal NaturalEffects_GraphMediation (if needed).
        """
        if adjustment_set is None:
            adjustment_set = 'auto'
        self.BlockedMediators = blocked_mediators
        if len(self.X) != 1:
            raise NotImplementedError("Currently only implemented for univariate effects (source).")
        if len(self.Y) != 1:
            raise NotImplementedError("Currently only implemented for univariate effects (target).")
        [source] = self.X
        [target] = self.Y
        self.MediationEstimator = NaturalEffects_GraphMediation(
            graph=self.graph, graph_type=self.graph_type, tau_max=self.tau_max,
            fit_setup=mixed_data_estimator, observations_data=dataframe,
            effect_source=source, effect_target=target,
            blocked_mediators=self.BlockedMediators, adjustment_set=adjustment_set, only_check_validity=False,
            fall_back_to_total_effect=use_mediation_impl_for_total_effect_fallback,
            _internal_provide_cfx=self, enable_dataframe_based_preprocessing=enable_dataframe_based_preprocessing)
        # return a NDE_Graph Estimator, but also remember it for predict_nde
        return self.MediationEstimator

    def predict_natural_direct_effect(self, reference_value_x, cf_intervention_value_x):
        """*After fitting* a natural direct effect, predict its value at a specific point.
        See also predict_natural_direct_effect_function.

        Parameters
        ----------
        reference_value_x : single value of same type as single sample for X (float, int, or bool)
            Reference-value to which X is set by intervention in the world seen by the (blocked) mediators.

        cf_intervention_value_x : single value of same type as single sample for X (float, int, or bool)
            Post-intervention-value to which X is set by intervention in the world seen by the effect (directly).

        Throws
        ------
        Raises and exception if parameters are not meaningful or if adjustment-set is not valid.

        Returns
        -------
        NDE : If Y is categorical -> np.array( # categories Y, 2 )
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.

        NDE : If Y is continuous -> float
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.
        """
        if self.MediationEstimator is None:
            raise Exception("Call fit_natural_direct_effect_x before using predict_natural_direct_effect.")
        return self.MediationEstimator.NDE(change_from=reference_value_x, change_to=cf_intervention_value_x)

    def predict_natural_direct_effect_function(self, min_x, max_x, cf_delta=0.5,
                                               steps=100, smoothing_gaussian_sigma_in_steps=5,
                                               normalize_by_delta=False):
        """Compute NDE as *smoothed* function

        Parameters
        ----------
        min_x : float
            Lower bound of interval on which reference-values for X are taken

        max_x : float
            Upper bound of interval on which reference-values for X are taken

        cf_delta : float
            The change from reference-value to effect-value (change_from=reference, change_to=ref + delta)

        steps : uint
            Number of intermediate values to compute in the interval [min_x, max_x]

        smoothing_gaussian_sigma_in_steps : uint
            The width of the Gauß-kernel used for smoothing, given in steps.

        normalize_by_delta : bool
            Normalize the effect by dividing by cf_delta.

        Throws
        ------
        Raises and exception if parameters are not meaningful, adjustment-set is not valid or
        normalization requested makes no sense (normalizing probabilites by delta).

        Returns
        -------
        NDE : If Y is categorical -> np.array( # steps, # categories Y, 2 )
            For each grid-point:
            The probabilities the categories of Y (after, before) changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.

        NDE : If Y is continuous -> np.array( # steps )
            For each grid-point:
            The change in the expectation-value of Y induced by changing the interventional value of X
            as "seen" by Y from change_from to change_to, while keeping (blocked) M as if X remained at change_from.
        """
        if self.MediationEstimator is None:
            raise Exception("Call fit_natural_direct_effect before using predict_natural_direct_effect_x.")
        return self.MediationEstimator.NDE_smoothed(min_x, max_x, cf_delta,
                                                    steps, smoothing_gaussian_sigma_in_steps,
                                                    normalize_by_delta)
